import numpy as np

import meshplex


def test_control_volume_centroids_line():
    X = np.array([0.0, 0.1, 0.4, 1.0])
    cells = np.array([[0, 1], [1, 2], [2, 3]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array([0.025, 0.15, 0.475, 0.85])

    print(mesh.control_volume_centroids)
    assert mesh.control_volume_centroids.shape == ref.shape
    assert np.all(
        np.abs(ref - mesh.control_volume_centroids) < np.abs(ref) * 1.0e-13 + 1.0e-13
    )


def test_control_volume_centroids_tri():
    points = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.Mesh(points, cells)

    ref = np.array([[0.25, 0.25], [2.0 / 3.0, 1.0 / 6.0], [1.0 / 6.0, 2.0 / 3.0]])

    print(mesh.control_volume_centroids)
    assert mesh.control_volume_centroids.shape == ref.shape
    assert np.all(
        np.abs(ref - mesh.control_volume_centroids) < np.abs(ref) * 1.0e-13 + 1.0e-13
    )


def test_tetra_simple():
    X = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2, 3]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array(
        [
            [0.25, 0.25, 0.25],
            [17 / 24, 1 / 48, 1 / 48],
            [1 / 48, 17 / 24, 1 / 48],
            [1 / 48, 1 / 48, 17 / 24],
        ]
    )
    print(mesh.control_volume_centroids)
    assert np.all(
        np.abs(mesh.control_volume_centroids - ref) < 1.0e-14 * np.abs(ref) + 1.0e-14
    )



# from helpers import download_mesh
# import meshplex
#
# import numpy as np
# import unittest
#
#
# class GradientTest(unittest.TestCase):
#
#     def setUp(self):
#         return
#
#     def _run_test(self, mesh):
#         num_nodes = len(mesh.points)
#         # Create function  2*x + 3*y.
#         a_x = 7.0
#         a_y = 3.0
#         a0 = 1.0
#         u = a_x * mesh.points[:, 0] + \
#             a_y * mesh.points[:, 1] + \
#             a0 * np.ones(num_nodes)
#         # Get the gradient analytically.
#         sol = np.empty((num_nodes, 2))
#         sol[:, 0] = a_x
#         sol[:, 1] = a_y
#         # Compute the gradient numerically.
#         grad_u = mesh.compute_gradient(u)
#
#         tol = 1.0e-13
#         for k in range(num_nodes):
#             self.assertAlmostEqual(grad_u[k][0], sol[k][0], delta=tol)
#             self.assertAlmostEqual(grad_u[k][1], sol[k][1], delta=tol)
#         return
#
#     def test_pacman(self):
#         filename = download_mesh(
#                 'pacman.vtk',
#                 '2da8ff96537f844a95a83abb48471b6a'
#                 )
#         mesh, _, _, _ = meshplex.read(filename)
#         self._run_test(mesh)
#         return
#
#
# if __name__ == '__main__':
#     unittest.main()



import matplotlib.pyplot as plt
import numpy as np

import meshplex


def _main():
    points = np.array([[0.0, 0.0], [1.0, 0.0], [0.3, 0.8]])
    # points = np.array([[0.0, 0.0], [1.0, 0.0], [0.5, np.sqrt(3) / 2]])
    cells = np.array([[0, 1, 2]])

    mesh = meshplex.Mesh(points, cells)

    lw = 5.0
    col = "0.6"

    ax = plt.gca()

    # circumcircle
    circle1 = plt.Circle(
        mesh.cell_circumcenters[0],
        mesh.cell_circumradius[0],
        color=col,
        fill=False,
        linewidth=lw,
    )
    ax.add_patch(circle1)

    # perpendicular bisectors
    for i, j in [[0, 1], [1, 2], [2, 0]]:
        m1 = (points[i] + points[j]) / 2
        v1 = m1 - mesh.cell_circumcenters[0]
        e1 = (
            mesh.cell_circumcenters[0]
            + v1 / np.linalg.norm(v1) * mesh.cell_circumradius[0]
        )
        plt.plot(
            [mesh.cell_circumcenters[0, 0], e1[0]],
            [mesh.cell_circumcenters[0, 1], e1[1]],
            col,
            linewidth=lw,
        )

    # heights
    for i, j, k in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
        p = points - points[i]
        v1 = p[j] / np.linalg.norm(p[j])
        m1 = points[i] + np.dot(p[k], v1) * v1
        plt.plot([points[k, 0], m1[0]], [points[k, 1], m1[1]], linewidth=lw, color=col)

    # # incircle
    # circle2 = plt.Circle(
    #     mesh.cell_incenters[0], mesh.inradius[0], color=col, fill=False, linewidth=lw
    # )
    # ax.add_patch(circle2)

    # # angle bisectors
    # for i, j, k in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
    #     p = points - points[i]
    #     v1 = p[j] / np.linalg.norm(p[j])
    #     v2 = p[k] / np.linalg.norm(p[k])
    #     alpha = np.arccos(np.dot(v1, v2))
    #     c = np.cos(alpha / 2)
    #     s = np.sin(alpha / 2)
    #     beta = np.linalg.norm(mesh.cell_incenters[0] - points[i]) + mesh.inradius[0]
    #     m1 = points[i] + np.dot([[c, -s], [s, c]], v1) * beta
    #     plt.plot(
    #         [points[i, 0], m1[0]],
    #         [points[i, 1], m1[1]],
    #         col,
    #         linewidth=lw,
    #         color=col,
    #     )

    # triangle
    plt.plot(
        [points[0, 0], points[1, 0], points[2, 0], points[0, 0]],
        [points[0, 1], points[1, 1], points[2, 1], points[0, 1]],
        color="#d62728",
        linewidth=lw,
    )

    ax.set_xlim(-0.1, 1.1)
    ax.set_ylim(-0.4, 0.9)
    ax.set_aspect("equal")
    plt.axis("off")
    plt.savefig("logo.svg", bbox_inches="tight", transparent=True)
    # plt.show()


if __name__ == "__main__":
    _main()



import math

import numpy as np

import meshplex


def test_heights_line():
    X = np.array([0.0, 0.1, 0.4, 1.0])
    cells = np.array([[0, 1], [1, 2], [2, 3]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array([0.1, 0.3, 0.6])
    print(mesh.cell_heights)
    assert np.all(np.abs(mesh.cell_heights - ref) < np.abs(ref) * 1.0e-13)


def test_heights_tri():
    # two triangles in 5D
    points = [
        [0.0, 0.0, 0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0, 0.0, 0.0],
        [1.0, 1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0, 0.0],
    ]
    cells = [[0, 1, 2], [0, 3, 2]]
    mesh = meshplex.Mesh(points, cells)

    ref = np.array(
        [
            [1.0, math.sqrt(0.5), 1.0],
            [1.0, math.sqrt(0.5), 1.0],
        ]
    ).T
    assert np.all(np.abs(mesh.cell_heights - ref) < np.abs(ref) * 1.0e-13)


def test_heights_tetra():
    # two triangles in 5D
    points = [
        [0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
    ]
    cells = [[0, 1, 2, 3]]
    mesh = meshplex.Mesh(points, cells)

    ref = np.array(
        [
            [math.sqrt(1 / 3), 1.0, 1.0, 1.0],
        ]
    ).T
    assert np.all(np.abs(mesh.cell_heights - ref) < np.abs(ref) * 1.0e-13)


if __name__ == "__main__":
    # test_heights_tri()
    test_heights_tetra()



import numpy as np

import meshplex


def test_circumcenters_line():
    pts = [0.0, 1.0, 3.0, 4.0]
    cells = [[0, 1], [1, 2], [2, 3]]
    mesh = meshplex.Mesh(pts, cells)
    print(mesh.cell_circumcenters)
    ref = [0.5, 2.0, 3.5]
    assert np.all(np.abs(mesh.cell_circumcenters - ref) < np.abs(ref) * 1.0e-13)

    print(mesh.cell_circumradius)
    ref = [0.5, 1.0, 0.5]
    assert np.all(np.abs(mesh.cell_circumradius - ref) < np.abs(ref) * 1.0e-13)


def test_circumcenters_tri():
    # two triangles in 5D
    points = [
        [0.0, 0.0, 0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0, 0.0, 0.0],
        [1.0, 1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0, 0.0],
    ]
    cells = [[0, 1, 2], [0, 3, 2]]
    mesh = meshplex.Mesh(points, cells)

    ref = [[0.5, 0.5, 0.0, 0.0, 0.0], [0.5, 0.5, 0.0, 0.0, 0.0]]
    assert np.all(
        np.abs(mesh.cell_circumcenters - ref) < np.abs(ref) * 1.0e-13 + 1.0e-13
    )

    print(mesh.cell_circumradius)
    ref = [np.sqrt(2) / 2, np.sqrt(2) / 2]
    assert np.all(np.abs(mesh.cell_circumradius - ref) < np.abs(ref) * 1.0e-13)


def test_circumcenters_tetra():
    points = [
        [0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
    ]
    cells = [[0, 1, 2, 3]]
    mesh = meshplex.Mesh(points, cells)

    ref = [[0.5, 0.5, 0.5]]
    print(mesh.cell_circumcenters)
    assert np.all(np.abs(mesh.cell_circumcenters - ref) < np.abs(ref) * 1.0e-13)

    print(mesh.cell_circumradius)
    ref = [np.sqrt(3) / 2]
    assert np.all(np.abs(mesh.cell_circumradius - ref) < np.abs(ref) * 1.0e-13)


def test_circumcenters_simplex5():
    points = [
        [0.0, 0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0, 1.0],
    ]
    cells = [[0, 1, 2, 3, 4]]
    mesh = meshplex.Mesh(points, cells)

    ref = [[0.5, 0.5, 0.5, 0.5]]
    print(mesh.cell_circumcenters)
    assert np.all(np.abs(mesh.cell_circumcenters - ref) < np.abs(ref) * 1.0e-13)

    print(mesh.cell_circumradius)
    ref = [1.0]
    assert np.all(np.abs(mesh.cell_circumradius - ref) < np.abs(ref) * 1.0e-13)


if __name__ == "__main__":
    # test_circumcenters_tri()
    test_circumcenters_tetra()



import pathlib

import meshplex

this_dir = pathlib.Path(__file__).resolve().parent


def test_pacman():
    mesh = meshplex.read(this_dir / "meshes" / "pacman.vtu")
    mesh = meshplex.MeshTri(mesh.points[:, :2], mesh.cells("points"))

    mesh._compute_cell_values()
    mesh._compute_cell_values(mask=[0, 12, 53, 54, 55])



import os
import pathlib
import platform
import tempfile

import meshio
import meshzoo
import numpy as np
import pytest

import meshplex

from ..helpers import assert_norms, is_near_equal, run

this_dir = pathlib.Path(__file__).resolve().parent


def _compute_polygon_area(pts):
    # shoelace formula
    return (
        np.abs(
            np.dot(pts[0], np.roll(pts[1], -1)) - np.dot(np.roll(pts[0], -1), pts[1])
        )
        / 2
    )


# The dtype restriction is because of np.bincount.
# See  <https://github.com/numpy/numpy/issues/17760> and
# <https://github.com/nschloe/meshplex/issues/90>.
cell_dtypes = []
cell_dtypes += [
    np.int32,
]
if platform.architecture()[0] == "64bit":
    cell_dtypes += [
        np.uint32,  # when numpy is fixed, this can go to all arches
        np.int64,
        # np.uint64  # depends on the numpy fix
    ]


@pytest.mark.parametrize("cells_dtype", cell_dtypes)
def test_unit_triangle(cells_dtype):
    points = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    cells = np.array([[0, 1, 2]], dtype=cells_dtype)
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-14

    # ce_ratios
    assert is_near_equal(mesh.ce_ratios.T, [0.0, 0.5, 0.5], tol)

    # control volumes
    assert is_near_equal(mesh.control_volumes, [0.25, 0.125, 0.125], tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5], tol)

    # circumcenters
    assert is_near_equal(mesh.cell_circumcenters, [0.5, 0.5], tol)

    # centroids
    assert is_near_equal(mesh.cell_centroids, [1.0 / 3.0, 1.0 / 3.0], tol)
    assert is_near_equal(mesh.cell_barycenters, [1.0 / 3.0, 1.0 / 3.0], tol)

    # control volume centroids
    print(mesh.control_volume_centroids)
    assert is_near_equal(
        mesh.control_volume_centroids,
        [[0.25, 0.25], [2.0 / 3.0, 1.0 / 6.0], [1.0 / 6.0, 2.0 / 3.0]],
        tol,
    )

    # incenter
    assert is_near_equal(
        mesh.cell_incenters, [[(2 - np.sqrt(2)) / 2, (2 - np.sqrt(2)) / 2]], tol
    )

    # circumcenter
    assert is_near_equal(mesh.cell_circumcenters, [[0.5, 0.5]], tol)

    assert mesh.num_delaunay_violations == 0

    assert mesh.genus == 0

    mesh.get_cell_mask()
    mesh.get_edge_mask()
    mesh.get_vertex_mask()

    # dummy subdomain marker test
    class Subdomain:
        is_boundary_only = False

        def is_inside(self, X):
            return np.ones(X.shape[1:], dtype=bool)

    cell_mask = mesh.get_cell_mask(Subdomain())
    assert np.sum(cell_mask) == 1

    # save
    _, filename = tempfile.mkstemp(suffix=".png")
    mesh.save(filename)
    os.remove(filename)
    _, filename = tempfile.mkstemp(suffix=".vtk")
    mesh.save(filename)
    os.remove(filename)


def test_regular_tri_additional_points():
    points = np.array(
        [
            [0.0, 3.4, 0.0],
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [3.3, 4.4, 0.0],
        ]
    )
    cells = np.array([[1, 2, 3]])
    mesh = meshplex.MeshTri(points, cells)

    assert np.array_equal(mesh.is_point_used, [False, True, True, True, False])
    assert np.array_equal(mesh.is_boundary_point, [False, True, True, True, False])
    assert np.array_equal(mesh.is_interior_point, [False, False, False, False, False])

    tol = 1.0e-14

    assert np.array_equal(mesh.cells("points"), [[1, 2, 3]])

    mesh.create_facets()
    assert np.array_equal(mesh.cells("edges"), [[2, 1, 0]])
    assert np.array_equal(mesh.edges["points"], [[1, 2], [1, 3], [2, 3]])

    # ce_ratios
    assert is_near_equal(mesh.ce_ratios.T, [0.0, 0.5, 0.5], tol)

    # control volumes
    assert is_near_equal(mesh.control_volumes, [0.0, 0.25, 0.125, 0.125, 0.0], tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5], tol)

    # circumcenters
    assert is_near_equal(mesh.cell_circumcenters, [0.5, 0.5, 0.0], tol)

    # Centroids.
    # Nans appear here as the some points aren't part of any cell and hence have no
    # control volume.
    cvc = mesh.control_volume_centroids
    assert np.all(np.isnan(cvc[0]))
    assert np.all(np.isnan(cvc[4]))
    assert is_near_equal(
        cvc[1:4],
        [[0.25, 0.25, 0.0], [2.0 / 3.0, 1.0 / 6.0, 0.0], [1.0 / 6.0, 2.0 / 3.0, 0.0]],
        tol,
    )
    assert mesh.num_delaunay_violations == 0


def test_regular_tri_order():
    points = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
    cells = np.array([[0, 1, 2]])

    mesh = meshplex.MeshTri(points, cells)
    assert all((mesh.cells("points") == [0, 1, 2]).flat)

    tol = 1.0e-14

    # ce_ratios
    assert is_near_equal(mesh.ce_ratios.T, [0.5, 0.0, 0.5], tol)

    # control volumes
    assert is_near_equal(mesh.control_volumes, [0.125, 0.25, 0.125], tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5], tol)

    # circumcenters
    assert is_near_equal(mesh.cell_circumcenters, [0.5, 0.5, 0.0], tol)

    # centroids
    assert is_near_equal(
        mesh.control_volume_centroids,
        [[1.0 / 6.0, 2.0 / 3.0, 0.0], [0.25, 0.25, 0.0], [2.0 / 3.0, 1.0 / 6.0, 0.0]],
        tol,
    )

    assert mesh.num_delaunay_violations == 0


@pytest.mark.parametrize("a", [1.0, 2.0])
def test_regular_tri2(a):
    points = (
        np.array(
            [
                [-0.5, -0.5 * np.sqrt(3.0), 0],
                [-0.5, +0.5 * np.sqrt(3.0), 0],
                [1, 0, 0],
            ]
        )
        / np.sqrt(3)
        * a
    )
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-14

    # ce_ratios
    val = 0.5 / np.sqrt(3.0)
    assert is_near_equal(mesh.ce_ratios, [val, val, val], tol)

    # control volumes
    vol = np.sqrt(3.0) / 4 * a**2
    assert is_near_equal(mesh.control_volumes, [vol / 3.0, vol / 3.0, vol / 3.0], tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [vol], tol)

    # circumcenters
    assert is_near_equal(mesh.cell_circumcenters, [0.0, 0.0, 0.0], tol)


# def test_degenerate_small0():
#     h = 1.0e-3
#     points = np.array([
#         [0, 0, 0],
#         [1, 0, 0],
#         [0.5, h, 0.0],
#         ])
#     cells = np.array([[0, 1, 2]])
#     mesh = meshplex.MeshTri(
#             points,
#             cells,
#             allow_negative_volumes=True
#             )

#     tol = 1.0e-14

#     # ce_ratios
#     alpha = 0.5 * h - 1.0 / (8*h)
#     beta = 1.0 / (4*h)
#     assertAlmostEqual(mesh.get_ce_ratios_per_edge()[0], alpha, delta=tol)
#     self.assertAlmostEqual(mesh.get_ce_ratios_per_edge()[1], beta, delta=tol)
#     self.assertAlmostEqual(mesh.get_ce_ratios_per_edge()[2], beta, delta=tol)

#     # control volumes
#     alpha1 = 0.0625 * (3*h - 1.0/(4*h))
#     alpha2 = 0.125 * (h + 1.0 / (4*h))
#     assert is_near_equal(
#         mesh.get_control_volumes(),
#         [alpha1, alpha1, alpha2],
#         tol
#         )

#     # cell volumes
#     self.assertAlmostEqual(mesh.cell_volumes[0], 0.5 * h, delta=tol)

#     # surface areas
#     edge_length = np.sqrt(0.5**2 + h**2)
#     # circumference = 1.0 + 2 * edge_length
#     alpha = 0.5 * (1.0 + edge_length)
#     self.assertAlmostEqual(mesh.surface_areas[0], alpha, delta=tol)
#     self.assertAlmostEqual(mesh.surface_areas[1], alpha, delta=tol)
#     self.assertAlmostEqual(mesh.surface_areas[2], edge_length, delta=tol)

#     # centroids
#     alpha = -41.666666669333345
#     beta = 0.58333199998399976
#      self.assertAlmostEqual(
#              mesh.centroids[0][0],
#              0.416668000016,
#              delta=tol
#              )
#     self.assertAlmostEqual(mesh.centroids[0][1], alpha, delta=tol)
#     self.assertAlmostEqual(mesh.centroids[0][2], 0.0, delta=tol)

#     self.assertAlmostEqual(mesh.centroids[1][0], beta, delta=tol)
#     self.assertAlmostEqual(mesh.centroids[1][1], alpha, delta=tol)
#     self.assertAlmostEqual(mesh.centroids[1][2], 0.0, delta=tol)

#     self.assertAlmostEqual(mesh.centroids[2][0], 0.5, delta=tol)
#     self.assertAlmostEqual(mesh.centroids[2][1], -41.666, delta=tol)
#     self.assertAlmostEqual(mesh.centroids[2][2], 0.0, delta=tol)

#     self.assertEqual(mesh.num_delaunay_violations, 0)


@pytest.mark.parametrize(
    "h",
    # TODO [1.0e0, 1.0e-1]
    [1.0e0],
)
def test_degenerate_small0b(h):
    points = np.array([[0, 0, 0], [1, 0, 0], [0.5, h, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells, sort_cells=True)  # test sort_cells, too

    tol = 1.0e-14

    # edge lengths
    el = np.sqrt(0.5**2 + h**2)
    assert is_near_equal(mesh.edge_lengths.T, [el, el, 1.0], tol)

    # ce_ratios
    ce0 = 0.5 / h * (h**2 - 0.25)
    ce12 = 0.25 / h
    assert is_near_equal(mesh.ce_ratios.T, [ce12, ce12, ce0], tol)

    # control volumes
    cv12 = 0.25 * (1.0**2 * ce0 + (0.25 + h**2) * ce12)
    cv0 = 0.5 * (0.25 + h**2) * ce12
    assert is_near_equal(mesh.control_volumes, [cv12, cv12, cv0], tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5 * h], tol)

    # circumcenters
    assert is_near_equal(mesh.cell_circumcenters, [0.5, 0.375, 0.0], tol)

    assert mesh.num_delaunay_violations == 0


# # TODO parametrize with flat boundary correction
# def test_degenerate_small0b_fcc():
#     h = 1.0e-3
#     points = np.array([[0, 0, 0], [1, 0, 0], [0.5, h, 0.0]])
#     cells = np.array([[0, 1, 2]])
#     mesh = meshplex.MeshTri(points, cells)
#
#     tol = 1.0e-14
#
#     # edge lengths
#     el = np.sqrt(0.5 ** 2 + h ** 2)
#     assert is_near_equal(mesh.edge_lengths.T, [el, el, 1.0], tol)
#
#     # ce_ratios
#     ce = h
#     assert is_near_equal(mesh.ce_ratios.T, [ce, ce, 0.0], tol)
#
#     # control volumes
#     cv = ce * el
#     alpha = 0.25 * el * cv
#     beta = 0.5 * h - 2 * alpha
#     assert is_near_equal(mesh.control_volumes, [alpha, alpha, beta], tol)
#
#     # cell volumes
#     assert is_near_equal(mesh.cell_volumes, [0.5 * h], tol)
#
#     # surface areas
#     g = np.sqrt((0.5 * el) ** 2 + (ce * el) ** 2)
#     alpha = 0.5 * el + g
#     beta = el + (1.0 - 2 * g)
#     assert is_near_equal(mesh.surface_areas, [alpha, alpha, beta], tol)
#
#     # centroids
#     centroids = mesh.control_volume_centroids
#     alpha = 1.0 / 6000.0
#     gamma = 0.00038888918518558031
#     assert is_near_equal(centroids[0], [0.166667, alpha, 0.0], tol)
#     assert is_near_equal(centroids[1], [0.833333, alpha, 0.0], tol)
#     assert is_near_equal(centroids[2], [0.5, gamma, 0.0], tol)

#     assert mesh.num_delaunay_violations == 0


@pytest.mark.parametrize("h, a", [(1.0e-3, 0.3)])
def test_degenerate_small1(h, a):
    points = np.array([[0, 0, 0], [1, 0, 0], [a, h, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-12

    # edge lengths
    el0 = np.sqrt((1.0 - a) ** 2 + h**2)
    el1 = np.sqrt(a**2 + h**2)
    el2 = 1.0
    assert is_near_equal(mesh.edge_lengths.T, [[el0, el1, el2]], tol)

    # ce_ratios
    ce0 = 0.5 * a / h
    ce1 = 0.5 * (1 - a) / h
    ce2 = 0.5 * (h - (1 - a) * a / h) / el2
    assert is_near_equal(mesh.ce_ratios[:, 0], [ce0, ce1, ce2], 1.0e-8)

    # # control volumes
    # cv1 = ce1 * el1
    # alpha1 = 0.25 * el1 * cv1
    # cv2 = ce2 * el2
    # alpha2 = 0.25 * el2 * cv2
    # beta = 0.5 * h - (alpha1 + alpha2)
    # assert is_near_equal(mesh.control_volumes, [alpha1, alpha2, beta], tol)
    # assert abs(sum(mesh.control_volumes) - 0.5 * h) < tol

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5 * h], tol)

    # # surface areas
    # b1 = np.sqrt((0.5 * el1) ** 2 + cv1 ** 2)
    # alpha0 = b1 + 0.5 * el1
    # b2 = np.sqrt((0.5 * el2) ** 2 + cv2 ** 2)
    # alpha1 = b2 + 0.5 * el2
    # total = 1.0 + el1 + el2
    # alpha2 = total - alpha0 - alpha1
    # assert is_near_equal(mesh.surface_areas, [alpha0, alpha1, alpha2], tol)

    assert mesh.num_delaunay_violations == 0


@pytest.mark.parametrize("h", [1.0e-2])
def test_degenerate_small2(h):
    points = np.array([[0, 0, 0], [1, 0, 0], [0.5, h, 0.0], [0.5, -h, 0.0]])
    cells = np.array([[0, 1, 2], [0, 1, 3]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-11

    # ce_ratios
    alpha = h - 1.0 / (4 * h)
    beta = 1.0 / (4 * h)
    assert is_near_equal(mesh.signed_circumcenter_distances, [alpha], tol)

    alpha2 = (h - 1.0 / (4 * h)) / 2
    assert is_near_equal(
        mesh.ce_ratios, [[beta, beta], [beta, beta], [alpha2, alpha2]], tol
    )

    # control volumes
    alpha1 = 0.125 * (3 * h - 1.0 / (4 * h))
    alpha2 = 0.125 * (h + 1.0 / (4 * h))
    assert is_near_equal(mesh.control_volumes, [alpha1, alpha1, alpha2, alpha2], tol)

    # circumcenters
    assert is_near_equal(
        mesh.cell_circumcenters, [[0.5, -12.495, 0.0], [0.5, +12.495, 0.0]], tol
    )

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [0.5 * h, 0.5 * h], tol)

    assert mesh.num_delaunay_violations == 1


def test_rectanglesmall():
    points = np.array(
        [[0.0, 0.0, 0.0], [10.0, 0.0, 0.0], [10.0, 1.0, 0.0], [0.0, 1.0, 0.0]]
    )
    cells = np.array([[0, 1, 2], [0, 2, 3]])

    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-14

    assert is_near_equal(mesh.signed_circumcenter_distances, [0.0], tol)

    assert is_near_equal(mesh.ce_ratios, [[5.0, 0.05], [0.0, 5.0], [0.05, 0.0]], tol)
    assert is_near_equal(mesh.control_volumes, [2.5, 2.5, 2.5, 2.5], tol)
    assert is_near_equal(mesh.cell_volumes, [5.0, 5.0], tol)
    assert mesh.num_delaunay_violations == 0


def test_pacman():
    mesh = meshplex.read(this_dir / ".." / "meshes" / "pacman.vtu")

    run(
        mesh,
        54.312974717523744,
        [1.9213504740523146, 0.07954185111555329],
        [403.5307055719196, 0.5512267577002408],
        [1.3816992621175055, 0.0443755870238773],
    )

    assert mesh.num_delaunay_violations == 0


def test_shell():
    points = np.array(
        [
            [+0.0, +0.0, +1.0],
            [+1.0, +0.0, +0.0],
            [+0.0, +1.0, +0.0],
            [-1.0, +0.0, +0.0],
            [+0.0, -1.0, +0.0],
        ]
    )
    cells = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 1, 4]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-14
    ce_ratios = 0.5 / np.sqrt(3.0) * np.ones((4, 3))
    assert is_near_equal(mesh.ce_ratios.T, ce_ratios, tol)

    cv = np.array([2.0, 1.0, 1.0, 1.0, 1.0]) / np.sqrt(3.0)
    assert is_near_equal(mesh.control_volumes, cv, tol)

    cell_vols = np.sqrt(3.0) / 2.0 * np.ones(4)
    assert is_near_equal(mesh.cell_volumes, cell_vols, tol)

    assert mesh.num_delaunay_violations == 0


def test_sphere():
    points, cells = meshzoo.icosa_sphere(5)
    mesh = meshplex.Mesh(points, cells)

    run(
        mesh,
        12.413437988936916,
        [0.7864027242108207, 0.05524648209283611],
        [128.70115197256447, 0.3605511489598192],
        [0.5593675314375034, 0.02963260270642986],
    )


def test_update_point_coordinates():
    mesh = meshio.read(this_dir / ".." / "meshes" / "pacman.vtu")
    assert np.all(np.abs(mesh.points[:, 2]) < 1.0e-15)

    mesh1 = meshplex.MeshTri(mesh.points, mesh.get_cells_type("triangle"))

    np.random.seed(123)
    X2 = mesh.points + 1.0e-2 * np.random.rand(*mesh.points.shape)
    mesh2 = meshplex.MeshTri(X2, mesh.get_cells_type("triangle"))

    mesh1.points = X2

    tol = 1.0e-12
    assert is_near_equal(mesh1.cell_volumes, mesh2.cell_volumes, tol)


def test_inradius():
    # 3-4-5 triangle
    points = np.array([[0.0, 0.0, 0.0], [3.0, 0.0, 0.0], [0.0, 4.0, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-15

    assert is_near_equal(mesh.cell_inradius, [1.0], tol)

    # 30-60-90 triangle
    a = 1.0
    points = np.array(
        [[0.0, 0.0, 0.0], [a / 2, 0.0, 0.0], [0.0, a / 2 * np.sqrt(3.0), 0.0]]
    )
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    assert is_near_equal(mesh.cell_inradius, [a / 4 * (np.sqrt(3) - 1)], tol)


def test_circumradius():
    # 3-4-5 triangle
    points = np.array([[0.0, 0.0, 0.0], [3.0, 0.0, 0.0], [0.0, 4.0, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-15

    assert is_near_equal(mesh.cell_circumradius, [2.5], tol)

    # 30-60-90 triangle
    a = 1.0
    points = np.array(
        [[0.0, 0.0, 0.0], [a / 2, 0.0, 0.0], [0.0, a / 2 * np.sqrt(3.0), 0.0]]
    )
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    assert is_near_equal(mesh.cell_circumradius, [a / 2], tol)


def test_quality():
    # 3-4-5 triangle
    points = np.array([[0.0, 0.0, 0.0], [3.0, 0.0, 0.0], [0.0, 4.0, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-15

    q = mesh.q_radius_ratio
    assert is_near_equal(q, 2 * mesh.cell_inradius / mesh.cell_circumradius, tol)

    # 30-60-90 triangle
    a = 1.0
    points = np.array(
        [[0.0, 0.0, 0.0], [a / 2, 0.0, 0.0], [0.0, a / 2 * np.sqrt(3.0), 0.0]]
    )
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    q = mesh.q_radius_ratio
    assert is_near_equal(q, 2 * mesh.cell_inradius / mesh.cell_circumradius, tol)


def test_angles():
    # 3-4-5 triangle
    points = np.array([[0.0, 0.0, 0.0], [3.0, 0.0, 0.0], [0.0, 4.0, 0.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    tol = 1.0e-14

    assert is_near_equal(
        mesh.angles,
        [[np.pi / 2], [np.arcsin(4.0 / 5.0)], [np.arcsin(3.0 / 5.0)]],
        tol,
    )

    # 30-60-90 triangle
    a = 1.0
    points = np.array(
        [[0.0, 0.0, 0.0], [a / 2, 0.0, 0.0], [0.0, a / 2 * np.sqrt(3.0), 0.0]]
    )
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    ic = mesh.angles / np.pi * 180
    assert is_near_equal(ic, [[90], [60], [30]], tol)


def test_flat_boundary():
    #
    #  3___________2
    #  |\_   2   _/|
    #  |  \_   _/  |
    #  | 3  \4/  1 |
    #  |   _/ \_   |
    #  | _/     \_ |
    #  |/    0    \|
    #  0-----------1
    #
    x = 0.4
    y = 0.5
    X = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [x, y, 0.0],
        ]
    )
    cells = np.array([[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4]])

    mesh = meshplex.MeshTri(X, cells)

    # Inspect the covolumes in left cell.
    edge_length = np.sqrt(x**2 + y**2)
    ref = np.array([edge_length, edge_length, 1.0])
    assert np.all(np.abs(mesh.edge_lengths[:, 3] - ref) < 1.0e-12)
    #
    alpha = 0.5 / x * y * np.sqrt(y**2 + x**2)
    beta = 0.5 / x * (x**2 - y**2)
    ref = [alpha, alpha, beta]
    covolumes = mesh.ce_ratios[:, 3] * mesh.edge_lengths[:, 3]
    assert np.all(np.abs(covolumes - ref) < 1.0e-12)

    #
    beta = np.sqrt(alpha**2 + 0.2**2 + 0.25**2)
    control_volume_corners = np.array(
        [
            mesh.cell_circumcenters[0][:2],
            mesh.cell_circumcenters[1][:2],
            mesh.cell_circumcenters[2][:2],
            mesh.cell_circumcenters[3][:2],
        ]
    )
    ref_area = _compute_polygon_area(control_volume_corners.T)

    assert np.abs(mesh.control_volumes[4] - ref_area) < 1.0e-12

    cv = np.zeros(X.shape[0])
    for edges, ce_ratios in zip(mesh.idx[1].T, mesh.ce_ratios.T):
        for i, ce in zip(edges, ce_ratios):
            ei = mesh.points[i[1]] - mesh.points[i[0]]
            cv[i] += 0.25 * ce * np.dot(ei, ei)

    assert np.all(np.abs(cv - mesh.control_volumes) < 1.0e-12 * cv)


def test_set_points():
    points = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    cells = np.array([[0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.set_points([0.1, 0.1], [0])
    ref = mesh.cell_volumes.copy()

    mesh2 = meshplex.MeshTri(mesh.points, mesh.cells("points"))
    assert np.all(np.abs(ref - mesh2.cell_volumes) < 1.0e-10)


def test_reference_vals_pacman():
    mesh = meshplex.read(this_dir / ".." / "meshes" / "pacman.vtu")
    mesh = meshplex.MeshTri(mesh.points[:, :2], mesh.cells("points"))

    assert_norms(
        mesh.points,
        [3.0544932927920363e03, 8.8106015937625088e01, 4.2500000000000000e00],
        1.0e-15,
    )
    assert_norms(
        mesh.half_edge_coords,
        [1.6674048509514694e03, 1.9463913181988705e01, 3.4718650853971766e-01],
        1.0e-15,
    )
    assert_norms(
        mesh.ei_dot_ei,
        [3.7884391635599366e02, 5.5826366101867908e00, 1.2739502744897091e-01],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_partitions,
        [5.4312974717523744e01, 5.6678942198355931e-01, 8.5844237283573006e-03],
        1.0e-12,
    )
    assert_norms(
        mesh.cell_centroids,
        [5.5718766084264298e03, 1.1728553160741399e02, 4.1694621840309081e00],
        1.0e-15,
    )
    assert_norms(
        mesh.edge_lengths,
        [1.3250455158127431e03, 1.9463913181988705e01, 3.5692440018716975e-01],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_volumes,
        [5.4312974717523744e01, 1.3816992621175055e00, 4.4375587023877297e-02],
        1.0e-15,
    )
    assert_norms(
        mesh.ce_ratios,
        [1.3499477445918124e03, 2.0088073714816950e01, 5.5122675770024077e-01],
        1.0e-14,
    )
    assert_norms(
        mesh.control_volumes,
        [5.4312974717523744e01, 1.9213504740523146e00, 7.9541851115553286e-02],
        1.0e-15,
    )
    assert_norms(
        mesh.control_volume_centroids,
        [3.0478135855839828e03, 8.7829558499197603e01, 4.1869842124121526e00],
        1.0e-15,
    )
    assert_norms(
        mesh.signed_cell_volumes,
        [5.4312974717523744e01, 1.3816992621175055e00, 4.4375587023877297e-02],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_circumcenters,
        [5.5720855984960217e03, 1.1729075391802718e02, 4.1780370020570583e00],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_circumradius,
        [2.5571964535497142e02, 6.5009888666474742e00, 1.8757161840809547e-01],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_incenters,
        [5.5715778346847819e03, 1.1727887700257899e02, 4.1655515539293466e00],
        1.0e-15,
    )
    assert_norms(
        mesh.cell_inradius,
        [1.2685029853822100e02, 3.2249724655140719e00, 9.1724742697552028e-02],
        1.0e-15,
    )
    assert_norms(
        mesh.q_radius_ratio,
        [1.5359568026022387e03, 3.9044827334140905e01, 9.9999895608618172e-01],
        1.0e-15,
    )



import numpy as np

import meshplex


def test_line():
    X = np.array([0.0, 0.1, 0.4, 1.0])
    cells = np.array([[0, 1], [1, 2], [2, 3]])
    mesh = meshplex.Mesh(X, cells)

    vol = np.sum(mesh.cell_volumes)
    assert np.all(np.abs(np.sum(mesh.control_volumes) - vol) < 1.0e-14 * np.abs(vol))

    ref = np.array([0.05, 0.2, 0.45, 0.3])
    assert np.all(np.abs(mesh.control_volumes - ref) < 1.0e-14 * np.abs(ref))


def test_tri_degen():
    X = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.5, 1.0e-2],
        ]
    )

    cells = np.array([[0, 1, 2]])
    mesh = meshplex.Mesh(X, cells)

    vol = np.sum(mesh.cell_volumes)
    assert np.all(np.abs(np.sum(mesh.control_volumes) - vol) < 1.0e-12 * np.abs(vol))

    ref = np.array([-1.560625, -1.560625, 3.12625])
    print(mesh.control_volumes)
    assert np.all(np.abs(mesh.control_volumes - ref) < 1.0e-14 * np.abs(ref))


def test_tri():
    X = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )

    cells = np.array([[0, 1, 2], [0, 2, 3]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array([0.25, 0.25, 0.25, 0.25])
    assert np.all(np.abs(mesh.control_volumes - ref) < 1.0e-14 * np.abs(ref))


def test_tetra_simple():
    X = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2, 3]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array([9.0, 1.0, 1.0, 1.0]) / 72
    mesh.control_volumes
    print()
    print(mesh.control_volumes)
    print(sum(mesh.control_volumes))
    print(ref)
    print(sum(ref))
    assert np.all(np.abs(mesh.control_volumes - ref) < 1.0e-14 * np.abs(ref) + 1.0e-14)


def test_tetra():
    X = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, -1.0],
            [0.0, 0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2, 3], [0, 1, 2, 4]])
    mesh = meshplex.Mesh(X, cells)

    ref = np.array([18.0, 2.0, 2.0, 1.0, 1.0]) / 72
    print(mesh.control_volumes)
    print(ref)
    assert np.all(np.abs(mesh.control_volumes - ref) < 1.0e-14 * np.abs(ref) + 1.0e-14)



import meshzoo
import pytest

import meshplex


def test_euler_characteristic():
    points = [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
    cells = [[0, 1, 2]]
    mesh = meshplex.MeshTri(points, cells)
    assert mesh.euler_characteristic == 1
    assert mesh.genus == 0

    points, cells = meshzoo.icosa_sphere(5)
    mesh = meshplex.MeshTri(points, cells)
    assert mesh.euler_characteristic == 2
    assert mesh.genus == 0

    points, cells = meshzoo.moebius(num_twists=1, nl=21, nw=6)
    mesh = meshplex.MeshTri(points, cells)
    assert mesh.euler_characteristic == 0
    with pytest.raises(RuntimeError):
        mesh.genus



import warnings

import npx
import numpy as np

from ._mesh import Mesh

__all__ = ["MeshTri"]


class MeshTri(Mesh):
    """Class for handling triangular meshes."""

    def __init__(self, points, cells, sort_cells=False):
        super().__init__(points, cells, sort_cells=sort_cells)

        assert self.n == 3

        # some backwards-compatibility fixes
        self.create_edges = super().create_facets
        self.compute_signed_cell_areas = super().compute_signed_cell_volumes

        self.boundary_edges = super().boundary_facets
        self.is_boundary_edge = super().is_boundary_facet

    @property
    def euler_characteristic(self):
        # number of vertices - number of edges + number of faces
        if self._cells_facets is None:
            self.create_facets()
        return (
            self.points.shape[0]
            - self.edges["points"].shape[0]
            + self.cells("points").shape[0]
        )

    @property
    def genus(self):
        # <https://math.stackexchange.com/a/4196497/36678>:
        #
        # chi = 2 - 2 * g - b
        #
        # g = 1 - (chi + b) / 2
        #
        # where b is the number of boundary components.
        if self._is_boundary_facet is None:
            self.create_facets()

        if np.any(self.is_boundary_edge):
            from scipy.sparse import coo_matrix
            from scipy.sparse.csgraph import connected_components

            boundary_edges = self.edges["points"][self.is_boundary_edge]

            n = np.max(boundary_edges) + 1
            val = np.ones(len(boundary_edges), dtype=int)
            M = coo_matrix((val, boundary_edges.T), shape=(n, n))
            # Unfortunately, connected_components() counts orphaned nodes as an
            # individual component. See <https://github.com/scipy/scipy/issues/14391>.
            _, cc = connected_components(M)
            num_cc = np.sum(np.bincount(cc) > 1)

        else:
            num_cc = 0

        chi = self.euler_characteristic
        if (chi + num_cc) % 2 != 0:
            raise RuntimeError("Non-integer genus. Is the mesh perhaps not orientable?")

        return 1 - (chi + num_cc) // 2

    @property
    def angles(self):
        """All angles in the triangle."""
        # The cosines of the angles are the negative dot products of the normalized
        # edges adjacent to the angle.
        norms = self.edge_lengths
        ei_dot_ej = self.ei_dot_ei - np.sum(self.ei_dot_ei, axis=0) / 2
        normalized_ei_dot_ej = np.array(
            [
                ei_dot_ej[0] / norms[1] / norms[2],
                ei_dot_ej[1] / norms[2] / norms[0],
                ei_dot_ej[2] / norms[0] / norms[1],
            ]
        )
        return np.arccos(-normalized_ei_dot_ej)

    #     def compute_gradient(self, u):
    #         '''Computes an approximation to the gradient :math:`\\nabla u` of a
    #         given scalar valued function :math:`u`, defined in the points.
    #         This is taken from
    #
    #            Discrete gradient method in solid mechanics,
    #            Lu, Jia and Qian, Jing and Han, Weimin,
    #            International Journal for Numerical Methods in Engineering,
    #            <https://doi.org/10.1002/nme.2187>.
    #         '''
    #         if self.cell_circumcenters is None:
    #             self.cell_circumcenters = self._circumcenters[-1]
    #
    #         if 'cells' not in self.edges:
    #             self.edges['cells'] = self.compute_edge_cells()
    #
    #         # This only works for flat meshes.
    #         assert (abs(self.points[:, 2]) < 1.0e-10).all()
    #         points2d = self.points[:, :2]
    #         cell_circumcenters2d = self.cell_circumcenters[:, :2]
    #
    #         num_points = len(points2d)
    #         assert len(u) == num_points
    #
    #         gradient = np.zeros((num_points, 2), dtype=u.dtype)
    #
    #         # Create an empty 2x2 matrix for the boundary points to hold the
    #         # edge correction ((17) in [1]).
    #         boundary_matrices = {}
    #         for point in self.get_vertices('boundary'):
    #             boundary_matrices[point] = np.zeros((2, 2))
    #
    #         for edge_gid, edge in enumerate(self.edges['cells']):
    #             # Compute edge length.
    #             point0 = self.edges['points'][edge_gid][0]
    #             point1 = self.edges['points'][edge_gid][1]
    #
    #             # Compute coedge length.
    #             if len(self.edges['cells'][edge_gid]) == 1:
    #                 # Boundary edge.
    #                 edge_midpoint = 0.5 * (
    #                         points2d[point0] +
    #                         points2d[point1]
    #                         )
    #                 cell0 = self.edges['cells'][edge_gid][0]
    #                 coedge_midpoint = 0.5 * (
    #                         cell_circumcenters2d[cell0] +
    #                         edge_midpoint
    #                         )
    #             elif len(self.edges['cells'][edge_gid]) == 2:
    #                 cell0 = self.edges['cells'][edge_gid][0]
    #                 cell1 = self.edges['cells'][edge_gid][1]
    #                 # Interior edge.
    #                 coedge_midpoint = 0.5 * (
    #                         cell_circumcenters2d[cell0] +
    #                         cell_circumcenters2d[cell1]
    #                         )
    #             else:
    #                 raise RuntimeError(
    #                         'Edge needs to have either one or two neighbors.'
    #                         )
    #
    #             # Compute the coefficient r for both contributions
    #             coeffs = self.ce_ratios[edge_gid] / \
    #                 self.control_volumes[self.edges['points'][edge_gid]]
    #
    #             # Compute R*_{IJ} ((11) in [1]).
    #             r0 = (coedge_midpoint - points2d[point0]) * coeffs[0]
    #             r1 = (coedge_midpoint - points2d[point1]) * coeffs[1]
    #
    #             diff = u[point1] - u[point0]
    #
    #             gradient[point0] += r0 * diff
    #             gradient[point1] -= r1 * diff
    #
    #             # Store the boundary correction matrices.
    #             edge_coords = points2d[point1] - points2d[point0]
    #             if point0 in boundary_matrices:
    #                 boundary_matrices[point0] += np.outer(r0, edge_coords)
    #             if point1 in boundary_matrices:
    #                 boundary_matrices[point1] += np.outer(r1, -edge_coords)
    #
    #         # Apply corrections to the gradients on the boundary.
    #         for k, value in boundary_matrices.items():
    #             gradient[k] = np.linalg.solve(value, gradient[k])
    #
    #         return gradient

    def compute_ncurl(self, vector_field):
        """Computes the n.dot.curl of a vector field over the mesh. While the vector
        field is point-based, the curl will be cell-based. The approximation is based on

        .. math::
            n\\cdot curl(F) = \\lim_{A\\to 0} |A|^{-1} \\rangle\\int_{dGamma}, F\\rangle dr;

        see https://en.wikipedia.org/wiki/Curl_(mathematics). Actually, to approximate
        the integral, one would only need the projection of the vector field onto the
        edges at the midpoint of the edges.
        """
        # Compute the projection of A on the edge at each edge midpoint. Take the
        # average of `vector_field` at the endpoints to get the approximate value at the
        # edge midpoint.
        A = 0.5 * np.sum(vector_field[self.idx[-1]], axis=0)
        # sum of <edge, A> for all three edges
        sum_edge_dot_A = np.einsum("ijk, ijk->j", self.half_edge_coords, A)

        # Get normalized vector orthogonal to triangle
        z = np.cross(self.half_edge_coords[0], self.half_edge_coords[1])

        # Now compute
        #
        #    curl = z / ||z|| * sum_edge_dot_A / |A|.
        #
        # Since ||z|| = 2*|A|, one can save a sqrt and do
        #
        #    curl = z * sum_edge_dot_A * 0.5 / |A|^2.
        #
        curl = z * (0.5 * sum_edge_dot_A / self.cell_volumes**2)[..., None]
        return curl

    def show_vertex(self, *args, **kwargs):
        """Show the mesh around a vertex (see plot_vertex())."""
        import matplotlib.pyplot as plt

        self.plot_vertex(*args, **kwargs)
        plt.show()
        plt.close()

    def plot_vertex(self, point_id, show_ce_ratio=True):
        """Plot the vicinity of a point and its covolume/edgelength ratio.

        :param point_id: Node ID of the point to be shown.
        :type point_id: int

        :param show_ce_ratio: If true, shows the ce_ratio of the point, too.
        :type show_ce_ratio: bool, optional
        """
        # Importing matplotlib takes a while, so don't do that at the header.
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.gca()
        plt.axis("equal")

        if self.edges is None:
            self.create_facets()

        # Find the edges that contain the vertex
        edge_gids = np.where((self.edges["points"] == point_id).any(axis=1))[0]
        # ... and plot them
        for point_ids in self.edges["points"][edge_gids]:
            x = self.points[point_ids]
            ax.plot(x[:, 0], x[:, 1], "k")

        # Highlight ce_ratios.
        if show_ce_ratio:
            # Find the cells that contain the vertex
            cell_ids = np.where((self.cells("points") == point_id).any(axis=1))[0]

            for cell_id in cell_ids:
                for edge_gid in self.cells("edges")[cell_id]:
                    if point_id not in self.edges["points"][edge_gid]:
                        continue
                    point_ids = self.edges["points"][edge_gid]
                    edge_midpoint = 0.5 * (
                        self.points[point_ids[0]] + self.points[point_ids[1]]
                    )
                    p = np.stack(
                        [self.cell_circumcenters[cell_id], edge_midpoint], axis=1
                    )
                    q = np.column_stack(
                        [
                            self.cell_circumcenters[cell_id],
                            edge_midpoint,
                            self.points[point_id],
                        ]
                    )
                    ax.fill(q[0], q[1], color="0.5")
                    ax.plot(p[0], p[1], color="0.7")
        return

    def flip_until_delaunay(self, tol=0.0, max_steps=100):
        """Flip edges until the mesh is fully Delaunay (up to `tol`)."""
        num_flips = 0
        assert tol >= 0.0
        # If all circumcenter-facet distances are positive, all cells are Delaunay.
        if np.all(self.circumcenter_facet_distances > -0.5 * tol):
            return num_flips

        # Now compute the boundary facet. A little more costly, but we'd have to do that
        # anyway. If all _interior_ circumcenter-facet distances are positive, all cells
        # are Delaunay.
        if np.all(
            self.circumcenter_facet_distances[~self.is_boundary_facet_local]
            > -0.5 * tol
        ):
            return num_flips

        step = 0

        is_flip_interior_facet = self.signed_circumcenter_distances < -tol

        while True:
            # Don't flip the edges which would flip into existing edges. This can
            # happen, for example, in triangular shell meshes in 3D, and leads to weird
            # behavior down the line. See
            # <https://github.com/nschloe/meshplex/issues/130>.
            flip_filter = np.ones(np.sum(is_flip_interior_facet), dtype=bool)
            #
            facets_cells_flip = self.facets_cells["interior"][:, is_flip_interior_facet]
            # facet_gids = facets_cells_flip[0]
            adj_cells = facets_cells_flip[1:3]
            lids = facets_cells_flip[3:5]
            expected_new_edges = np.array(
                [
                    self.cells("points")[adj_cells[0], lids[0]],
                    self.cells("points")[adj_cells[1], lids[1]],
                ]
            ).T
            expected_new_edges = np.sort(expected_new_edges, axis=1)
            #
            # This isin() call can be quite costly since we're checking against _all_
            # existing edges.
            already_exists = npx.isin_rows(expected_new_edges, self.facets["points"])
            flip_filter &= ~already_exists
            #
            # Check if some flips would lead to the same flipped edge.
            _, inv = npx.unique_rows(expected_new_edges, return_inverse=True)
            is_unique = np.zeros(len(expected_new_edges), dtype=bool)
            is_unique[inv] = True
            flip_filter &= is_unique
            # apply the filter
            is_flip_interior_facet[is_flip_interior_facet] &= flip_filter

            step += 1
            if not np.any(is_flip_interior_facet):
                break

            if step > max_steps:
                break

            interior_facets_cells = self.facets_cells["interior"][1:3].T
            adj_cells = interior_facets_cells[is_flip_interior_facet].T

            # Check if there are cells for which more than one facet needs to be
            # flipped. For those, only flip one facet, namely that with the smaller
            # (more negative) circumcenter_facet_distance.
            cell_gids, num_flips_per_cell = np.unique(adj_cells, return_counts=True)
            multiflip_cell_gids = cell_gids[num_flips_per_cell > 1]
            while np.any(num_flips_per_cell > 1):
                for cell_gid in multiflip_cell_gids:
                    facet_gids = self.cells("facets")[cell_gid]
                    is_interior_facet = self.is_interior_facet[facet_gids]
                    idx = self.facets_cells_idx[facet_gids[is_interior_facet]]
                    k = np.argmin(self.signed_circumcenter_distances[idx])
                    is_flip_interior_facet[idx] = False
                    is_flip_interior_facet[idx[k]] = True

                adj_cells = interior_facets_cells[is_flip_interior_facet].T
                cell_gids, num_flips_per_cell = np.unique(adj_cells, return_counts=True)
                multiflip_cell_gids = cell_gids[num_flips_per_cell > 1]

            # actually perform the flips
            self.flip_interior_facets(is_flip_interior_facet)
            num_flips += np.sum(is_flip_interior_facet)

            is_flip_interior_facet_old = is_flip_interior_facet.copy()
            # Check which edges need to be flipped next.
            is_flip_interior_facet = self.signed_circumcenter_distances < -tol

            # Don't flip edges which have just been flipped. (This can happen due to
            # round-off errors.)
            is_flip_interior_facet[is_flip_interior_facet_old] = False

        is_negative = self.signed_circumcenter_distances < -tol
        num_nondelaunay_facets = np.sum(is_negative)
        if num_nondelaunay_facets > 0:
            dists = self.signed_circumcenter_distances[is_negative]
            warnings.warn(
                f"After {step} edge flip steps, "
                + f"there are {num_nondelaunay_facets} remaining non-Delaunay facets. "
                f"The signed circumcenter distances are {dists.tolist()}. "
                + "This can happen due to round-off errors or "
                + "to prevent non-manifold edges in shell meshes."
            )

        return num_flips

    def flip_interior_facets(self, is_flip_interior_facet):
        facets_cells_flip = self.facets_cells["interior"][:, is_flip_interior_facet]
        facet_gids = facets_cells_flip[0]
        adj_cells = facets_cells_flip[1:3]
        lids = facets_cells_flip[3:5]

        #        3                   3
        #        A                   A
        #       /|\                 / \
        #     3/ | \2             3/   \2
        #     /  |  \             /  1  \
        #   0/ 0 |   \1   ==>   0/_______\1
        #    \   | 1 /           \       /
        #     \  |  /             \  0  /
        #     0\ | /1             0\   /1
        #       \|/                 \ /
        #        V                   V
        #        2                   2
        #
        v = np.array(
            [
                self.cells("points")[adj_cells[0], lids[0]],
                self.cells("points")[adj_cells[1], lids[1]],
                self.cells("points")[adj_cells[0], (lids[0] + 1) % 3],
                self.cells("points")[adj_cells[0], (lids[0] + 2) % 3],
            ]
        )

        # This must be computed before the points are reset
        equal_orientation = (
            self.cells("points")[adj_cells[0], (lids[0] + 1) % 3]
            == self.cells("points")[adj_cells[1], (lids[1] + 2) % 3]
        )

        # Set up new cells->points relationships.
        # Make sure that positive/negative area orientation is preserved. This is
        # especially important for signed area computations: In a mesh of all positive
        # areas, you don't want a negative area appear after a facet flip.
        self.cells("points")[adj_cells[0]] = v[[0, 2, 1]].T
        self.cells("points")[adj_cells[1]] = v[[0, 1, 3]].T

        # Set up new facet->points relationships.
        self.facets["points"][facet_gids] = np.sort(v[[0, 1]], axis=0).T

        # Set up new cells->facets relationships.
        previous_facets = self.cells("facets")[adj_cells].copy()  # TODO need copy?
        # Do the neighboring cells have equal orientation (both point sets
        # clockwise/counterclockwise)?
        #
        # facets as in the above ascii art
        i0 = np.ones(equal_orientation.shape[0], dtype=int)
        i0[~equal_orientation] = 2
        i1 = np.ones(equal_orientation.shape[0], dtype=int)
        i1[equal_orientation] = 2
        e = [
            np.choose((lids[0] + 2) % 3, previous_facets[0].T),
            np.choose((lids[1] + i0) % 3, previous_facets[1].T),
            np.choose((lids[1] + i1) % 3, previous_facets[1].T),
            np.choose((lids[0] + 1) % 3, previous_facets[0].T),
        ]
        # The order here is tightly coupled to self.cells("points") above
        self.cells("facets")[adj_cells[0]] = np.column_stack([e[1], facet_gids, e[0]])
        self.cells("facets")[adj_cells[1]] = np.column_stack([e[2], e[3], facet_gids])

        # update is_boundary_facet_local
        for k in range(3):
            self.is_boundary_facet_local[k, adj_cells] = self.is_boundary_facet[
                self.cells("facets")[adj_cells, k]
            ]

        # Update the facet->cells relationship. We need to update facets_cells info for
        # all five facets.
        # First update the flipped facet; it's always interior.
        idx = self.facets_cells_idx[facet_gids]
        # cell ids
        self.facets_cells["interior"][1, idx] = adj_cells[0]
        self.facets_cells["interior"][2, idx] = adj_cells[1]
        # local facet ids; see self.cells("facets")
        self.facets_cells["interior"][3, idx] = 1
        self.facets_cells["interior"][4, idx] = 2
        #
        # Now handle the four surrounding facets
        conf = [
            # The data is:
            # (1) facet id
            # (2) previous adjacent cell (adj_cells[0] or adj_cells[1])
            # (3) new adjacent cell (adj_cells[0] or adj_cells[1])
            # (4) local facet index in the new adjacent cell
            (e[0], 0, 0, 2),
            (e[1], 1, 0, 0),
            (e[2], 1, 1, 0),
            (e[3], 0, 1, 1),
        ]
        for facet, prev_adj_idx, new__adj_idx, new_local_facet_index in conf:
            prev_adj = adj_cells[prev_adj_idx]
            new__adj = adj_cells[new__adj_idx]
            idx = self.facets_cells_idx[facet]
            # boundary...
            is_boundary = self.is_boundary_facet[facet]
            idx_bou = idx[is_boundary]
            prev_adjacent = prev_adj[is_boundary]
            new__adjacent = new__adj[is_boundary]
            # The assertion just makes sure we're doing the right thing. It should never
            # trigger.
            assert np.all(prev_adjacent == self.facets_cells["boundary"][1, idx_bou])
            self.facets_cells["boundary"][1, idx_bou] = new__adjacent
            self.facets_cells["boundary"][2, idx_bou] = new_local_facet_index
            # ...or interior?
            prev_adjacent = prev_adj[~is_boundary]
            new__adjacent = new__adj[~is_boundary]
            idx_int = idx[~is_boundary]
            # Interior facets have two neighboring cells in no particular order. Find
            # out if the adj_cell if the flipped facet comes first or second.
            is_first = prev_adjacent == self.facets_cells["interior"][1, idx_int]
            # The following is just a safety net. We could as well take ~is_first.
            is_secnd = prev_adjacent == self.facets_cells["interior"][2, idx_int]
            assert np.all(np.logical_xor(is_first, is_secnd))
            # actually set the data
            idx_first = idx_int[is_first]
            self.facets_cells["interior"][1, idx_first] = new__adjacent[is_first]
            self.facets_cells["interior"][3, idx_first] = new_local_facet_index
            # likewise for when the cell appears in the second column
            idx_secnd = idx_int[~is_first]
            self.facets_cells["interior"][2, idx_secnd] = new__adjacent[~is_first]
            self.facets_cells["interior"][4, idx_secnd] = new_local_facet_index

        # Schedule the cell ids for data updates
        update_cell_ids = np.unique(adj_cells.T.flat)
        # Same for facet ids
        update_facet_gids = self.cells("facets")[update_cell_ids].flat
        facet_cell_idx = self.facets_cells_idx[update_facet_gids]
        update_interior_facet_ids = np.unique(
            facet_cell_idx[self.is_interior_facet[update_facet_gids]]
        )

        self._update_cell_values(update_cell_ids, update_interior_facet_ids)

    def _update_cell_values(self, cell_ids, interior_facet_ids):
        """Updates all sorts of cell information for the given cell IDs."""
        # update idx
        for j in range(1, self.n - 1):
            m = len(self.idx[j - 1])
            r = np.arange(m)
            k = np.array([np.roll(r, -i) for i in range(1, m)])
            self.idx[j][..., cell_ids] = self.idx[j - 1][..., cell_ids][k]

        # update most of the cell-associated values
        self._compute_cell_values(cell_ids)

        if self._signed_cell_volumes is not None:
            self._signed_cell_volumes[cell_ids] = self.compute_signed_cell_volumes(
                cell_ids
            )

        if self._is_boundary_cell is not None:
            self._is_boundary_cell[cell_ids] = np.any(
                self.is_boundary_facet_local[:, cell_ids], axis=0
            )

        if self._cell_centroids is not None:
            self._cell_centroids[cell_ids] = self.compute_cell_centroids(cell_ids)

        # update the signed circumcenter distances for all interior_facet_ids
        if self._signed_circumcenter_distances is not None:
            self._signed_circumcenter_distances[interior_facet_ids] = 0.0
            facet_gids = self.interior_facets[interior_facet_ids]
            adj_cells = self.facets_cells["interior"][1:3, interior_facet_ids].T

            for i in [0, 1]:
                is_facet = np.array(
                    [
                        self.cells("facets")[adj_cells[:, i]][:, k] == facet_gids
                        for k in range(3)
                    ]
                )
                # assert np.all(np.sum(is_facet, axis=0) == 1)
                for k in range(3):
                    self._signed_circumcenter_distances[
                        interior_facet_ids[is_facet[k]]
                    ] += self._circumcenter_facet_distances[
                        k, adj_cells[is_facet[k], i]
                    ]

        # TODO update those values
        self._control_volumes = None
        self._ce_ratios = None
        self._cv_centroids = None



import numpy as np


def grp_start_len(a):
    """Given a sorted 1D input array `a`, e.g., [0 0, 1, 2, 3, 4, 4, 4], this routine
    returns the indices where the blocks of equal integers start and how long the blocks
    are.
    """
    # https://stackoverflow.com/a/50394587/353337
    m = np.concatenate([[True], a[:-1] != a[1:], [True]])
    idx = np.flatnonzero(m)
    return idx[:-1], np.diff(idx)


def _dot(a, n):
    """Dot product, preserve the leading n dimensions."""
    # einsum is faster if the tail survives, e.g., ijk,ijk->jk.
    # <https://gist.github.com/nschloe/8bc015cc1a9e5c56374945ddd711df7b>
    # TODO reorganize the data?
    assert n <= len(a.shape)
    # Would use -1 as second argument, but <https://github.com/numpy/numpy/issues/18519>
    b = a.reshape(*a.shape[:n], np.prod(a.shape[n:]).astype(int))
    return np.einsum("...i,...i->...", b, b)


def _multiply(a, b, n):
    """Multiply the along the first n dimensions of a and b. For example,
    a.shape == (5,6,3), b.shape == (5, 6), n = 2, will return an array c of a.shape with
    c[i,j,k] = a[i,j,k] * b[i,j].
    """
    aa = a.reshape(np.prod(a.shape[:n]), *a.shape[n:])
    bb = b.reshape(np.prod(b.shape[:n]), *b.shape[n:])
    cc = (aa.T * bb).T
    c = cc.reshape(*a.shape)
    return c



from __future__ import annotations

import math
import pathlib
import warnings

import meshio
import npx
import numpy as np
from numpy.typing import ArrayLike, NDArray

from ._exceptions import MeshplexError
from ._helpers import _dot, _multiply, grp_start_len

__all__ = ["Mesh"]


class Mesh:
    def __init__(self, points, cells, sort_cells: bool = False):
        points = np.asarray(points)
        cells = np.asarray(cells)

        if sort_cells:
            # Sort cells, first every row, then the rows themselves. This helps in many
            # downstream applications, e.g., when constructing linear systems with the
            # cells/edges. (When converting to CSR format, the I/J entries must be
            # sorted.) Don't use cells.sort(axis=1) to avoid
            # ```
            # ValueError: sort array is read-only
            # ```
            cells = np.sort(cells, axis=1)
            cells = cells[cells[:, 0].argsort()]

        # assert len(points.shape) <= 2, f"Illegal point coordinates shape {points.shape}"
        assert len(cells.shape) == 2, f"Illegal cells shape {cells.shape}"
        self.n = cells.shape[1]

        # Assert that all vertices are used.
        # If there are vertices which do not appear in the cells list, this
        # ```
        # uvertices, uidx = np.unique(cells, return_inverse=True)
        # cells = uidx.reshape(cells.shape)
        # points = points[uvertices]
        # ```
        # helps.
        # is_used = np.zeros(len(points), dtype=bool)
        # is_used[cells] = True
        # assert np.all(is_used), "There are {} dangling points in the mesh".format(
        #     np.sum(~is_used)
        # )

        self._points = np.asarray(points)
        # prevent accidental override of parts of the array
        self._points.setflags(write=False)

        # Initialize the idx hierarchy. The first entry, idx[0], is the cells->points
        # relationship, shape [3, numcells] for triangles and [4, numcells] for
        # tetrahedra. idx[1] is the (half-)facet->points to relationship, shape [2, 3,
        # numcells] for triangles and [3, 4, numcells] for tetrahedra, for example. The
        # indexing is chosen such the point idx[0][k] is opposite of the facet idx[1][:,
        # k]. This indexing keeps going until idx[-1] is of shape [2, 3, ..., numcells].
        self.idx = [np.asarray(cells).T]
        for _ in range(1, self.n - 1):
            m = len(self.idx[-1])
            r = np.arange(m)
            k = np.array([np.roll(r, -i) for i in range(1, m)])
            self.idx.append(self.idx[-1][k])

        self._is_point_used = None

        self._is_boundary_facet = None
        self._is_boundary_facet_local = None
        self.facets = None
        self._boundary_facets = None
        self._interior_facets = None
        self._is_interior_point = None
        self._is_boundary_point = None
        self._is_boundary_cell = None
        self._cells_facets = None

        self.subdomains = {}

        self._reset_point_data()

    def _reset_point_data(self):
        """Reset all data that changes when point coordinates changes."""
        self._half_edge_coords = None
        self._ei_dot_ei = None
        self._cell_centroids = None
        self._volumes = None
        self._integral_x = None
        self._signed_cell_volumes = None
        self._circumcenters = None
        self._cell_circumradii = None
        self._cell_heights = None
        self._ce_ratios = None
        self._cell_partitions = None
        self._control_volumes = None
        self._signed_circumcenter_distances = None
        self._circumcenter_facet_distances = None

        self._cv_centroids = None
        self._cvc_cell_mask = None
        self._cv_cell_mask = None

    def __repr__(self):
        name = {
            2: "line",
            3: "triangle",
            4: "tetra",
        }[self.cells("points").shape[1]]
        num_points = len(self.points)
        num_cells = len(self.cells("points"))
        string = f"<meshplex {name} mesh, {num_points} points, {num_cells} cells>"
        return string

    # prevent overriding points without adapting the other mesh data
    @property
    def points(self) -> np.ndarray:
        return self._points

    @points.setter
    def points(self, new_points: ArrayLike):
        new_points = np.asarray(new_points)
        assert new_points.shape == self._points.shape
        self._points = new_points
        # reset all computed values
        self._reset_point_data()

    def set_points(self, new_points: ArrayLike, idx=slice(None)):
        self.points.setflags(write=True)
        self.points[idx] = new_points
        self.points.setflags(write=False)
        self._reset_point_data()

    def cells(self, which) -> NDArray[np.int_]:
        if which == "points":
            return self.idx[0].T
        elif which == "facets":
            assert self._cells_facets is not None
            return self._cells_facets

        assert which == "edges"
        assert self.n == 3
        assert self._cells_facets is not None
        return self._cells_facets

    @property
    def half_edge_coords(self) -> NDArray[np.float_]:
        if self._half_edge_coords is None:
            self._compute_cell_values()
        assert self._half_edge_coords is not None
        return self._half_edge_coords

    @property
    def ei_dot_ei(self) -> NDArray[np.int_]:
        if self._ei_dot_ei is None:
            self._compute_cell_values()
        assert self._ei_dot_ei is not None
        return self._ei_dot_ei

    @property
    def cell_heights(self) -> NDArray[np.float_]:
        if self._cell_heights is None:
            self._compute_cell_values()
        assert self._cell_heights is not None
        return self._cell_heights

    @property
    def edge_lengths(self) -> NDArray[np.float_]:
        if self._volumes is None:
            self._compute_cell_values()
        assert self._volumes is not None
        return self._volumes[0]

    @property
    def facet_areas(self) -> NDArray[np.float_]:
        if self.n == 2:
            assert self.facets is not None
            return np.ones(len(self.facets["points"]))
        if self._volumes is None:
            self._compute_cell_values()
        assert self._volumes is not None
        return self._volumes[-2]

    @property
    def cell_volumes(self) -> NDArray[np.float_]:
        if self._volumes is None:
            self._compute_cell_values()
        assert self._volumes is not None
        return self._volumes[-1]

    @property
    def cell_circumcenters(self) -> NDArray[np.float_]:
        """Get the center of the circumsphere of each cell."""
        if self._circumcenters is None:
            self._compute_cell_values()
        assert self._circumcenters is not None
        return self._circumcenters[-1]

    @property
    def cell_circumradius(self) -> NDArray[np.float_]:
        """Get the circumradii of all cells"""
        if self._cell_circumradii is None:
            self._compute_cell_values()
        assert self._cell_circumradii is not None
        return self._cell_circumradii

    @property
    def cell_partitions(self) -> NDArray[np.float_]:
        """Each simplex can be subdivided into parts that a closest to each corner.
        This method gives those parts, like ce_ratios associated with each edge.
        """
        if self._cell_partitions is None:
            self._compute_cell_values()
        assert self._cell_partitions is not None
        return self._cell_partitions

    @property
    def circumcenter_facet_distances(self) -> NDArray[np.float_]:
        if self._circumcenter_facet_distances is None:
            self._compute_cell_values()
        assert self._circumcenter_facet_distances is not None
        return self._circumcenter_facet_distances

    def get_control_volume_centroids(self, cell_mask=None):
        """The centroid of any volume V is given by

        .. math::
          c = \\int_V x / \\int_V 1.

        The denominator is the control volume. The numerator can be computed by making
        use of the fact that the control volume around any vertex is composed of right
        triangles, two for each adjacent cell.

        Optionally disregard the contributions from particular cells. This is useful,
        for example, for temporarily disregarding flat cells on the boundary when
        performing Lloyd mesh optimization.
        """
        if self._cv_centroids is None or np.any(cell_mask != self._cvc_cell_mask):
            if self._integral_x is None:
                self._compute_cell_values()

            if cell_mask is None:
                idx = Ellipsis
            else:
                cell_mask = np.asarray(cell_mask)
                assert cell_mask.dtype == bool
                assert cell_mask.shape == (self.idx[-1].shape[-1],)
                # Use ":" for the first n-1 dimensions, then cell_mask
                idx = tuple((self.n - 1) * [slice(None)] + [~cell_mask])

            # TODO this can be improved by first summing up all components per cell
            integral_p = npx.sum_at(
                self._integral_x[idx], self.idx[-1][idx], len(self.points)
            )

            # Divide by the control volume
            cv = self.get_control_volumes(cell_mask)
            self._cv_centroids = (integral_p.T / cv).T
            self._cvc_cell_mask = cell_mask

        return self._cv_centroids

    @property
    def control_volume_centroids(self):
        return self.get_control_volume_centroids()

    @property
    def ce_ratios(self) -> NDArray[np.float_]:
        """The covolume-edgelength ratios."""
        # There are many ways for computing the ratio of the covolume and the edge
        # length. For triangles, for example, there is
        #
        #   ce_ratios = -<ei, ej> / cell_volume / 4,
        #
        # for tetrahedra,
        #
        #   zeta = (
        #       + ei_dot_ej[0, 2] * ei_dot_ej[3, 5] * ei_dot_ej[5, 4]
        #       + ei_dot_ej[0, 1] * ei_dot_ej[3, 5] * ei_dot_ej[3, 4]
        #       + ei_dot_ej[1, 2] * ei_dot_ej[3, 4] * ei_dot_ej[4, 5]
        #       + self.ei_dot_ej[0] * self.ei_dot_ej[1] * self.ei_dot_ej[2]
        #       ).
        #
        # Since we have detailed cell partitions at hand, though, the easiest and
        # fastest is via those.
        if self._ce_ratios is None:
            self._ce_ratios = (
                self.cell_partitions[0] / self.ei_dot_ei * 2 * (self.n - 1)
            )

        return self._ce_ratios

    @property
    def signed_circumcenter_distances(self):
        if self._signed_circumcenter_distances is None:
            if self._cells_facets is None:
                self.create_facets()

            self._signed_circumcenter_distances = npx.sum_at(
                self.circumcenter_facet_distances.T,
                self.cells("facets"),
                self.facets["points"].shape[0],
            )[self.is_interior_facet]

        return self._signed_circumcenter_distances

    def _compute_cell_values(self, mask=None):
        """Computes the volumes of all edges, facets, cells etc. in the mesh. It starts
        off by computing the (squared) edge lengths, then complements the edge with one
        vertex to form face. It computes an orthogonal basis of the face (with modified
        Gram-Schmidt), and from that gets the height of all faces. From this, the area
        of the face is computed. Then, it complements again to form the 3-simplex,
        again forms an orthogonal basis with Gram-Schmidt, and so on.
        """
        if mask is None:
            mask = slice(None)

        e = self.points[self.idx[-1][..., mask]]
        e0 = e[0]
        diff = e[1] - e[0]

        orthogonal_basis = np.array([diff])

        volumes2 = [_dot(diff, self.n - 1)]

        circumcenters = [0.5 * (e[0] + e[1])]

        vv = _dot(diff, self.n - 1)
        circumradii2 = 0.25 * vv
        sqrt_vv = np.sqrt(vv)
        lmbda = 0.5 * sqrt_vv

        sumx = np.array(e + circumcenters[-1])

        partitions = 0.5 * np.array([sqrt_vv, sqrt_vv])

        norms2 = np.array(volumes2)
        for kk, idx in enumerate(self.idx[:-1][::-1]):
            # Use the orthogonal bases of all sides to get a vector `v` orthogonal to
            # the side, pointing towards the additional point `p0`.
            p0 = self.points[idx][:, mask]
            v = p0 - e0
            # modified gram-schmidt
            for w, w_dot_w in zip(orthogonal_basis, norms2):
                w_dot_v = np.einsum("...k,...k->...", w, v)
                # Compute <w, v> / <w, w>, but don't set the output value where w==0.
                # The value remains uninitialized and gets canceled out in the next
                # iteration when multiplied by w.
                alpha = np.divide(w_dot_v, w_dot_w, where=w_dot_w > 0.0, out=w_dot_v)
                v -= _multiply(w, alpha, self.n - 1 - kk)

            vv = np.einsum("...k,...k->...", v, v)

            # Form the orthogonal basis for the next iteration by choosing one side
            # `k0`. <https://gist.github.com/nschloe/3922801e200cf82aec2fb53c89e1c578>
            # shows that it doesn't make a difference which point-facet combination we
            # choose.
            k0 = 0
            e0 = e0[k0]
            orthogonal_basis = np.row_stack([orthogonal_basis[:, k0], [v[k0]]])
            norms2 = np.row_stack([norms2[:, k0], [vv[k0]]])

            # The squared volume is the squared volume of the face times the squared
            # height divided by (n+1) ** 2.
            volumes2.append(volumes2[-1][0] * vv[k0] / (kk + 2) ** 2)

            # get the distance to the circumcenter; used in cell partitions and
            # circumcenter/-radius computation
            c = circumcenters[-1]

            p0c2 = _dot(p0 - c, self.n - 1 - kk)
            # Be a bit careful here. sigma and lmbda can be negative. Also make sure
            # that the values aren't nan when they should be inf (for degenerate
            # simplices, i.e., vv == 0).
            a = 0.5 * (p0c2 - circumradii2)
            sqrt_vv = np.sqrt(vv)
            with warnings.catch_warnings():
                # silence division-by-0 warnings
                # Happens for degenerate cells (sqrt(vv) == 0), and this case is
                # supported by meshplex. The values lmbda and sigma will just be +-inf.
                warnings.simplefilter("ignore", category=RuntimeWarning)
                lmbda = a / sqrt_vv
                sigma_k0 = a[k0] / vv[k0]

            # circumcenter, squared circumradius
            # <https://math.stackexchange.com/a/4064749/36678>
            lmbda2_k0 = sigma_k0 * a[k0]
            circumradii2 = lmbda2_k0 + circumradii2[k0]
            with warnings.catch_warnings():
                # Similar as above: The multiplicattion `v * sigma` correctly produces
                # nans for degenerate cells.
                warnings.simplefilter("ignore", category=RuntimeWarning)
                circumcenters.append(
                    c[k0] + _multiply(v[k0], sigma_k0, self.n - 2 - kk)
                )

            sumx += circumcenters[-1]

            # cell partitions
            partitions *= lmbda / (kk + 2)

        # The integral of x,
        #
        #   \\int_V x,
        #
        # over all atomic wedges, i.e., areas cornered by a point, an edge midpoint, and
        # the subsequent circumcenters.
        # The integral of any linear function over a triangle is the average of the
        # values of the function in each of the three corners, times the area of the
        # triangle.
        integral_x = _multiply(sumx, partitions / self.n, self.n)

        if np.all(mask == slice(None)):
            # set new values
            self._ei_dot_ei = volumes2[0]
            self._half_edge_coords = diff
            self._volumes = [np.sqrt(v2) for v2 in volumes2]
            self._circumcenter_facet_distances = lmbda
            self._cell_heights = sqrt_vv
            self._cell_circumradii = np.sqrt(circumradii2)
            self._circumcenters = circumcenters
            self._cell_partitions = partitions
            self._integral_x = integral_x
        else:
            # update existing values
            assert self._ei_dot_ei is not None
            self._ei_dot_ei[:, mask] = volumes2[0]

            assert self._half_edge_coords is not None
            self._half_edge_coords[:, mask] = diff

            assert self._volumes is not None
            for k in range(len(self._volumes)):
                self._volumes[k][..., mask] = np.sqrt(volumes2[k])

            assert self._circumcenter_facet_distances is not None
            self._circumcenter_facet_distances[..., mask] = lmbda

            assert self._cell_heights is not None
            self._cell_heights[..., mask] = sqrt_vv

            assert self._cell_circumradii is not None
            self._cell_circumradii[mask] = np.sqrt(circumradii2)

            assert self._circumcenters is not None
            for k in range(len(self._circumcenters)):
                self._circumcenters[k][..., mask, :] = circumcenters[k]

            assert self._cell_partitions is not None
            self._cell_partitions[..., mask] = partitions

            assert self._integral_x is not None
            self._integral_x[..., mask, :] = integral_x

    @property
    def signed_cell_volumes(self):
        """Signed volumes of an n-simplex in nD."""
        if self._signed_cell_volumes is None:
            self._signed_cell_volumes = self.compute_signed_cell_volumes()
        return self._signed_cell_volumes

    def compute_signed_cell_volumes(self, idx=slice(None)):
        """Signed volume of a simplex in nD. Note that signing only makes sense for
        n-simplices in R^n.
        """
        n = self.points.shape[1]
        assert (
            self.n == self.points.shape[1] + 1
        ), f"Signed areas only make sense for n-simplices in in nD. Got {n}D points."
        if self.n == 3:
            # On <https://stackoverflow.com/q/50411583/353337>, we have a number of
            # alternatives computing the oriented area, but it's fastest with the
            # half-edges.
            x = self.half_edge_coords
            assert x is not None
            return (x[0, idx, 1] * x[2, idx, 0] - x[0, idx, 0] * x[2, idx, 1]) / 2

        # https://en.wikipedia.org/wiki/Simplex#Volume
        cp = self.points[self.cells("points")]
        # append 1s
        cp1 = np.concatenate([cp, np.ones(cp.shape[:-1] + (1,))], axis=-1)

        # There appears to be no canonical convention when it comes to the sign
        # <https://math.stackexchange.com/q/4209203/36678>. With the below choice, the
        # area of 1D simplices [a, b] with a < b are positive, and the common ordering
        # of tetrahedra (as in VTK, for example) is positive.
        sign = -1 if n % 2 == 1 else 1
        return sign * np.linalg.det(cp1) / math.factorial(n)

    def compute_cell_centroids(self, idx=slice(None)):
        return np.sum(self.points[self.cells("points")[idx]], axis=1) / self.n

    @property
    def cell_centroids(self) -> NDArray[np.float_]:
        """The centroids (barycenters, midpoints of the circumcircles) of all
        simplices."""
        if self._cell_centroids is None:
            self._cell_centroids = self.compute_cell_centroids()
        return self._cell_centroids

    cell_barycenters = cell_centroids

    @property
    def cell_incenters(self) -> NDArray[np.float_]:
        """Get the midpoints of the inspheres."""
        # https://en.wikipedia.org/wiki/Incenter#Barycentric_coordinates
        # https://math.stackexchange.com/a/2864770/36678
        abc = self.facet_areas / np.sum(self.facet_areas, axis=0)
        return np.einsum("ij,jik->jk", abc, self.points[self.cells("points")])

    @property
    def cell_inradius(self) -> NDArray[np.float_]:
        """Get the inradii of all cells"""
        # See <https://mathworld.wolfram.com/Incircle.html>.
        # https://en.wikipedia.org/wiki/Tetrahedron#Inradius
        return (self.n - 1) * self.cell_volumes / np.sum(self.facet_areas, axis=0)

    @property
    def is_point_used(self) -> NDArray[np.bool_]:
        # Check which vertices are used.
        # If there are vertices which do not appear in the cells list, this
        # ```
        # uvertices, uidx = np.unique(cells, return_inverse=True)
        # cells = uidx.reshape(cells.shape)
        # points = points[uvertices]
        # ```
        # helps.
        if self._is_point_used is None:
            self._is_point_used = np.zeros(len(self.points), dtype=bool)
            self._is_point_used[self.cells("points")] = True
        return self._is_point_used

    def write(
        self,
        filename: str,
        point_data=None,
        cell_data=None,
        field_data=None,
    ):
        if self.points.shape[1] == 2:
            n = len(self.points)
            a = np.ascontiguousarray(np.column_stack([self.points, np.zeros(n)]))
        else:
            a = self.points

        if self.cells("points").shape[1] == 3:
            cell_type = "triangle"
        else:
            assert (
                self.cells("points").shape[1] == 4
            ), "Only triangles/tetrahedra supported"
            cell_type = "tetra"

        meshio.Mesh(
            a,
            {cell_type: self.cells("points")},
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data,
        ).write(filename)

    def get_vertex_mask(self, subdomain=None):
        if subdomain is None:
            # https://stackoverflow.com/a/42392791/353337
            return slice(None)
        if subdomain not in self.subdomains:
            self._mark_vertices(subdomain)
        return self.subdomains[subdomain]["vertices"]

    def get_edge_mask(self, subdomain=None):
        """Get faces which are fully in subdomain."""
        if subdomain is None:
            # https://stackoverflow.com/a/42392791/353337
            return slice(None)

        if subdomain not in self.subdomains:
            self._mark_vertices(subdomain)

        # A face is inside if all its edges are in.
        # An edge is inside if all its points are in.
        is_in = self.subdomains[subdomain]["vertices"][self.idx[-1]]
        # Take `all()` over the first index
        is_inside = np.all(is_in, axis=tuple(range(1)))

        if subdomain.is_boundary_only:
            # Filter for boundary
            is_inside = is_inside & self.is_boundary_facet

        return is_inside

    def get_face_mask(self, subdomain):
        """Get faces which are fully in subdomain."""
        if subdomain is None:
            # https://stackoverflow.com/a/42392791/353337
            return slice(None)

        if subdomain not in self.subdomains:
            self._mark_vertices(subdomain)

        # A face is inside if all its edges are in.
        # An edge is inside if all its points are in.
        is_in = self.subdomains[subdomain]["vertices"][self.idx[-1]]
        # Take `all()` over all axes except the last two (face_ids, cell_ids).
        n = len(is_in.shape)
        is_inside = np.all(is_in, axis=tuple(range(n - 2)))

        if subdomain.is_boundary_only:
            # Filter for boundary
            is_inside = is_inside & self.is_boundary_facet_local

        return is_inside

    def get_cell_mask(self, subdomain=None):
        if subdomain is None:
            # https://stackoverflow.com/a/42392791/353337
            return slice(None)

        if subdomain.is_boundary_only:
            # There are no boundary cells
            return np.array([])

        if subdomain not in self.subdomains:
            self._mark_vertices(subdomain)

        is_in = self.subdomains[subdomain]["vertices"][self.idx[-1]]
        # Take `all()` over all axes except the last one (cell_ids).
        n = len(is_in.shape)
        return np.all(is_in, axis=tuple(range(n - 1)))

    def _mark_vertices(self, subdomain):
        """Mark faces/edges which are fully in subdomain."""
        if subdomain is None:
            is_inside = np.ones(len(self.points), dtype=bool)
        else:
            is_inside = subdomain.is_inside(self.points.T).T

            if subdomain.is_boundary_only:
                # if the boundary hasn't been computed yet, this can take a little
                # moment
                is_inside &= self.is_boundary_point

        self.subdomains[subdomain] = {"vertices": is_inside}

    def create_facets(self):
        """Set up facet->point and facet->cell relations."""
        if self.n == 2:
            # Too bad that the need a specializaiton here. Could be avoided if the
            # idx hierarchy would be of shape (1,2,...,n), not (2,...,n), but not sure
            # if that's worth the change.
            idx = self.idx[0].flatten()
        else:
            idx = self.idx[1]
            idx = idx.reshape(idx.shape[0], -1)

        # Sort the columns to make it possible for `unique()` to identify individual
        # facets.
        idx = np.sort(idx, axis=0).T
        a_unique, inv, cts = npx.unique_rows(
            idx, return_inverse=True, return_counts=True
        )

        if np.any(cts > 2):
            num_weird_edges = np.sum(cts > 2)
            msg = (
                f"Found {num_weird_edges} facets with more than two neighboring cells. "
                "Something is not right."
            )
            # check if cells are identical, list them
            _, inv, cts = npx.unique_rows(
                np.sort(self.cells("points")), return_inverse=True, return_counts=True
            )
            if np.any(cts > 1):
                msg += " The following cells are equal:\n"
                for multiple_idx in np.where(cts > 1)[0]:
                    msg += str(np.where(inv == multiple_idx)[0])
            raise MeshplexError(msg)

        self._is_boundary_facet_local = (cts[inv] == 1).reshape(self.idx[0].shape)
        self._is_boundary_facet = cts == 1

        self.facets = {"points": a_unique}
        # cell->facets relationship
        self._cells_facets = inv.reshape(self.n, -1).T

        if self.n == 3:
            self.edges = self.facets
            self._facets_cells = None
            self._facets_cells_idx = None
        elif self.n == 4:
            self.faces = self.facets

    @property
    def is_boundary_facet_local(self) -> NDArray[np.bool_]:
        if self._is_boundary_facet_local is None:
            self.create_facets()
        assert self._is_boundary_facet_local is not None
        return self._is_boundary_facet_local

    @property
    def is_boundary_facet(self) -> NDArray[np.bool_]:
        if self._is_boundary_facet is None:
            self.create_facets()
        assert self._is_boundary_facet is not None
        return self._is_boundary_facet

    @property
    def is_interior_facet(self) -> NDArray[np.bool_]:
        return ~self.is_boundary_facet

    @property
    def is_boundary_cell(self):
        if self._is_boundary_cell is None:
            assert self.is_boundary_facet_local is not None
            self._is_boundary_cell = np.any(self.is_boundary_facet_local, axis=0)
        return self._is_boundary_cell

    @property
    def boundary_facets(self):
        if self._boundary_facets is None:
            self._boundary_facets = np.where(self.is_boundary_facet)[0]
        return self._boundary_facets

    @property
    def interior_facets(self):
        if self._interior_facets is None:
            self._interior_facets = np.where(~self.is_boundary_facet)[0]
        return self._interior_facets

    @property
    def is_boundary_point(self):
        if self._is_boundary_point is None:
            self._is_boundary_point = np.zeros(len(self.points), dtype=bool)
            # it's a little weird that we have to special-case n==2 here
            i = 0 if self.n == 2 else 1
            self._is_boundary_point[
                self.idx[i][..., self.is_boundary_facet_local]
            ] = True
        return self._is_boundary_point

    @property
    def is_interior_point(self):
        if self._is_interior_point is None:
            self._is_interior_point = self.is_point_used & ~self.is_boundary_point
        return self._is_interior_point

    @property
    def facets_cells(self):
        if self._facets_cells is None:
            self._compute_facets_cells()
        return self._facets_cells

    def _compute_facets_cells(self):
        """This creates edge->cells relations. While it's not necessary for many
        applications, it sometimes does come in handy, for example for mesh
        manipulation.
        """
        if self.facets is None:
            self.create_facets()

        # num_edges = len(self.edges["points"])
        # count = np.bincount(self.cells("edges").flat, minlength=num_edges)

        # <https://stackoverflow.com/a/50395231/353337>
        edges_flat = self.cells("edges").flat
        idx_sort = np.argsort(edges_flat)
        sorted_edges = edges_flat[idx_sort]
        idx_start, count = grp_start_len(sorted_edges)

        # count is redundant with is_boundary/interior_edge
        assert np.all((count == 1) == self.is_boundary_facet)
        assert np.all((count == 2) == self.is_interior_facet)

        idx_start_count_1 = idx_start[self.is_boundary_facet]
        idx_start_count_2 = idx_start[self.is_interior_facet]
        res1 = idx_sort[idx_start_count_1]
        res2 = idx_sort[np.array([idx_start_count_2, idx_start_count_2 + 1])]

        edge_id_boundary = sorted_edges[idx_start_count_1]
        edge_id_interior = sorted_edges[idx_start_count_2]

        # It'd be nicer if we could organize the data differently, e.g., as a structured
        # array or as a dict. Those possibilities are slower, unfortunately, for some
        # operations in remove_cells() (and perhaps elsewhere).
        # <https://github.com/numpy/numpy/issues/17850>
        self._facets_cells = {
            # rows:
            #  0: edge id
            #  1: cell id
            #  2: local edge id (0, 1, or 2)
            "boundary": np.array([edge_id_boundary, res1 // 3, res1 % 3]),
            # rows:
            #  0: edge id
            #  1: cell id 0
            #  2: cell id 1
            #  3: local edge id 0 (0, 1, or 2)
            #  4: local edge id 1 (0, 1, or 2)
            "interior": np.array([edge_id_interior, *(res2 // 3), *(res2 % 3)]),
        }

        self._facets_cells_idx = None

    @property
    def facets_cells_idx(self):
        if self._facets_cells_idx is None:
            if self._facets_cells is None:
                self._compute_facets_cells()
            assert self.is_boundary_facet is not None
            # For each edge, store the index into the respective edge array.
            num_edges = len(self.facets["points"])
            self._facets_cells_idx = np.empty(num_edges, dtype=int)
            num_b = np.sum(self.is_boundary_facet)
            num_i = np.sum(self.is_interior_facet)
            self._facets_cells_idx[self.facets_cells["boundary"][0]] = np.arange(num_b)
            self._facets_cells_idx[self.facets_cells["interior"][0]] = np.arange(num_i)
        return self._facets_cells_idx

    def remove_dangling_points(self):
        """Remove all points which aren't part of an array"""
        is_part_of_cell = np.zeros(self.points.shape[0], dtype=bool)
        is_part_of_cell[self.cells("points").flat] = True

        new_point_idx = np.cumsum(is_part_of_cell) - 1

        self._points = self._points[is_part_of_cell]

        for k in range(len(self.idx)):
            self.idx[k] = new_point_idx[self.idx[k]]

        if self._control_volumes is not None:
            self._control_volumes = self._control_volumes[is_part_of_cell]

        if self._cv_centroids is not None:
            self._cv_centroids = self._cv_centroids[is_part_of_cell]

        if self.facets is not None:
            self.facets["points"] = new_point_idx[self.facets["points"]]

        if self._is_interior_point is not None:
            self._is_interior_point = self._is_interior_point[is_part_of_cell]

        if self._is_boundary_point is not None:
            self._is_boundary_point = self._is_boundary_point[is_part_of_cell]

        if self._is_point_used is not None:
            self._is_point_used = self._is_point_used[is_part_of_cell]

        return np.sum(~is_part_of_cell)

    @property
    def q_radius_ratio(self):
        """Ratio of incircle and circumcircle ratios times (n-1). ("Normalized shape
        ratio".) Is 1 for the equilateral simplex, and is often used a quality measure
        for the cell.
        """
        # There are other sensible possibilities of defining cell quality, e.g.:
        #   * inradius to longest edge
        #   * shortest to longest edge
        #   * minimum dihedral angle
        #   * ...
        # See
        # <http://eidors3d.sourceforge.net/doc/index.html?eidors/meshing/calc_mesh_quality.html>.
        if self.n == 3:
            # q = 2 * r_in / r_out
            #   = (-a+b+c) * (+a-b+c) * (+a+b-c) / (a*b*c),
            #
            # where r_in is the incircle radius and r_out the circumcircle radius
            # and a, b, c are the edge lengths.
            a, b, c = self.edge_lengths
            return (-a + b + c) * (a - b + c) * (a + b - c) / (a * b * c)

        return (self.n - 1) * self.cell_inradius / self.cell_circumradius

    def remove_cells(self, remove_array: ArrayLike):
        """Remove cells and take care of all the dependent data structures. The input
        argument `remove_array` can be a boolean array or a list of indices.
        """
        # Although this method doesn't compute anything new, the reorganization of the
        # data structure is fairly expensive. This is mostly due to the fact that mask
        # copies like `a[mask]` take long if `a` is large, even if `mask` is True almost
        # everywhere.
        # Keep an eye on <https://stackoverflow.com/q/65035280/353337> for possible
        # workarounds.
        remove_array = np.asarray(remove_array)
        if len(remove_array) == 0:
            return 0

        if remove_array.dtype == int:
            keep = np.ones(len(self.cells("points")), dtype=bool)
            keep[remove_array] = False
        else:
            assert remove_array.dtype == bool
            keep = ~remove_array

        assert len(keep) == len(self.cells("points")), "Wrong length of index array."

        if np.all(keep):
            return 0

        # handle facet; this is a bit messy
        if self._cells_facets is not None:
            # updating the boundary data is a lot easier with facets_cells
            if self._facets_cells is None:
                self._compute_facets_cells()

            # Set facet to is_boundary_facet_local=True if it is adjacent to a removed
            # cell.
            facet_ids = self.cells("facets")[~keep].flatten()
            # only consider interior facets
            facet_ids = facet_ids[self.is_interior_facet[facet_ids]]
            idx = self.facets_cells_idx[facet_ids]
            cell_id = self.facets_cells["interior"][1:3, idx].T
            local_facet_id = self.facets_cells["interior"][3:5, idx].T
            self._is_boundary_facet_local[local_facet_id, cell_id] = True
            # now remove the entries corresponding to the removed cells
            self._is_boundary_facet_local = self._is_boundary_facet_local[:, keep]

            if self._is_boundary_cell is not None:
                self._is_boundary_cell[cell_id] = True
                self._is_boundary_cell = self._is_boundary_cell[keep]

            # update facets_cells
            keep_b_ec = keep[self.facets_cells["boundary"][1]]
            keep_i_ec0, keep_i_ec1 = keep[self.facets_cells["interior"][1:3]]
            # move ec from interior to boundary if exactly one of the two adjacent cells
            # was removed

            keep_i_0 = keep_i_ec0 & ~keep_i_ec1
            keep_i_1 = keep_i_ec1 & ~keep_i_ec0
            self._facets_cells["boundary"] = np.array(
                [
                    # facet id
                    np.concatenate(
                        [
                            self._facets_cells["boundary"][0, keep_b_ec],
                            self._facets_cells["interior"][0, keep_i_0],
                            self._facets_cells["interior"][0, keep_i_1],
                        ]
                    ),
                    # cell id
                    np.concatenate(
                        [
                            self._facets_cells["boundary"][1, keep_b_ec],
                            self._facets_cells["interior"][1, keep_i_0],
                            self._facets_cells["interior"][2, keep_i_1],
                        ]
                    ),
                    # local facet id
                    np.concatenate(
                        [
                            self._facets_cells["boundary"][2, keep_b_ec],
                            self._facets_cells["interior"][3, keep_i_0],
                            self._facets_cells["interior"][4, keep_i_1],
                        ]
                    ),
                ]
            )

            keep_i = keep_i_ec0 & keep_i_ec1

            # this memory copy isn't too fast
            self._facets_cells["interior"] = self._facets_cells["interior"][:, keep_i]

            num_facets_old = len(self.facets["points"])
            adjacent_facets, counts = np.unique(
                self.cells("facets")[~keep].flat, return_counts=True
            )
            # remove facet entirely either if 2 adjacent cells are removed or if it is a
            # boundary facet and 1 adjacent cells are removed
            is_facet_removed = (counts == 2) | (
                (counts == 1) & self._is_boundary_facet[adjacent_facets]
            )

            # set the new boundary facet
            self._is_boundary_facet[adjacent_facets[~is_facet_removed]] = True
            # Now actually remove the facets. This includes a reindexing.
            assert self._is_boundary_facet is not None
            keep_facets = np.ones(len(self._is_boundary_facet), dtype=bool)
            keep_facets[adjacent_facets[is_facet_removed]] = False

            # make sure there is only facets["points"], not facets["cells"] etc.
            assert self.facets is not None
            assert len(self.facets) == 1
            self.facets["points"] = self.facets["points"][keep_facets]
            self._is_boundary_facet = self._is_boundary_facet[keep_facets]

            # update facet and cell indices
            self._cells_facets = self.cells("facets")[keep]
            new_index_facets = np.arange(num_facets_old) - np.cumsum(~keep_facets)
            self._cells_facets = new_index_facets[self.cells("facets")]
            num_cells_old = len(self.cells("points"))
            new_index_cells = np.arange(num_cells_old) - np.cumsum(~keep)

            # this takes fairly long
            ec = self._facets_cells
            ec["boundary"][0] = new_index_facets[ec["boundary"][0]]
            ec["boundary"][1] = new_index_cells[ec["boundary"][1]]
            ec["interior"][0] = new_index_facets[ec["interior"][0]]
            ec["interior"][1:3] = new_index_cells[ec["interior"][1:3]]

            # simply set those to None; their reset is cheap
            self._facets_cells_idx = None
            self._boundary_facets = None
            self._interior_facets = None

        for k in range(len(self.idx)):
            self.idx[k] = self.idx[k][..., keep]

        if self._volumes is not None:
            for k in range(len(self._volumes)):
                self._volumes[k] = self._volumes[k][..., keep]

        if self._ce_ratios is not None:
            self._ce_ratios = self._ce_ratios[:, keep]

        if self._half_edge_coords is not None:
            self._half_edge_coords = self._half_edge_coords[:, keep]

        if self._ei_dot_ei is not None:
            self._ei_dot_ei = self._ei_dot_ei[:, keep]

        if self._cell_centroids is not None:
            self._cell_centroids = self._cell_centroids[keep]

        if self._circumcenters is not None:
            for k in range(len(self._circumcenters)):
                self._circumcenters[k] = self._circumcenters[k][..., keep, :]

        if self._cell_partitions is not None:
            self._cell_partitions = self._cell_partitions[..., keep]

        if self._signed_cell_volumes is not None:
            self._signed_cell_volumes = self._signed_cell_volumes[keep]

        if self._integral_x is not None:
            self._integral_x = self._integral_x[..., keep, :]

        if self._circumcenter_facet_distances is not None:
            self._circumcenter_facet_distances = self._circumcenter_facet_distances[
                ..., keep
            ]

        # TODO These could also be updated, but let's implement it when needed
        self._signed_circumcenter_distances = None
        self._control_volumes = None
        self._cv_cell_mask = None
        self._cv_centroids = None
        self._cvc_cell_mask = None
        self._is_point_used = None
        self._is_interior_point = None
        self._is_boundary_point = None

        return np.sum(~keep)

    def remove_boundary_cells(self, criterion):
        """Helper method for removing cells along the boundary.
        The input criterion is a callback that must return an array of length
        `sum(mesh.is_boundary_cell)`.

        This helps, for example, in the following scenario.
        When points are moving around, flip_until_delaunay() makes sure the mesh remains
        a Delaunay mesh. This does not work on boundaries where very flat cells can
        still occur or cells may even 'invert'. (The interior point moves outside.) In
        this case, the boundary cell can be removed, and the newly outward node is made
        a boundary node."""
        num_removed = 0
        while True:
            num_boundary_cells = np.sum(self.is_boundary_cell)
            crit = criterion(self.is_boundary_cell)

            if ~np.any(crit):
                break

            if not isinstance(crit, np.ndarray) or crit.shape != (num_boundary_cells,):
                raise ValueError(
                    "criterion() callback must return a Boolean NumPy array "
                    f"of shape {(num_boundary_cells,)}, got {crit.shape}."
                )

            idx = self.is_boundary_cell.copy()
            idx[idx] = crit
            n = self.remove_cells(idx)
            num_removed += n
            if n == 0:
                break
        return num_removed

    def remove_duplicate_cells(self):
        sorted_cells = np.sort(self.cells("points"))
        _, inv, cts = npx.unique_rows(
            sorted_cells, return_inverse=True, return_counts=True
        )

        remove = np.zeros(len(self.cells("points")), dtype=bool)
        for k in np.where(cts > 1)[0]:
            rem = inv == k
            # don't remove first occurrence
            first_idx = np.where(rem)[0][0]
            rem[first_idx] = False
            remove |= rem

        return self.remove_cells(remove)

    def get_control_volumes(self, cell_mask: ArrayLike | None = None) -> np.ndarray:
        """The control volumes around each vertex. Optionally disregard the
        contributions from particular cells. This is useful, for example, for
        temporarily disregarding flat cells on the boundary when performing Lloyd mesh
        optimization.
        """
        if cell_mask is not None:
            cell_mask = np.asarray(cell_mask)

        if self._cv_centroids is None or np.any(cell_mask != self._cvc_cell_mask):
            # Sum up the contributions according to how self.idx is constructed.
            # roll = np.array([np.roll(np.arange(kk + 3), -i) for i in range(1, kk + 3)])
            # vols = npx.sum_at(vols, roll, kk + 3)
            # v = self.cell_partitions[..., idx]

            if cell_mask is None:
                idx = slice(None)
            else:
                idx = ~cell_mask

            # TODO this can be improved by first summing up all components per cell
            self._control_volumes = npx.sum_at(
                self.cell_partitions[..., idx],
                self.idx[-1][..., idx],
                len(self.points),
            )

            self._cv_cell_mask = cell_mask

        assert self._control_volumes is not None
        return self._control_volumes

    control_volumes = property(get_control_volumes)

    @property
    def is_delaunay(self):
        return self.num_delaunay_violations == 0

    @property
    def num_delaunay_violations(self):
        """Number of interior facets where the Delaunay condition is violated."""
        # Delaunay violations are present exactly on the interior facets where the
        # signed circumcenter distance is negative. Count those.
        return np.sum(self.signed_circumcenter_distances < 0.0)

    @property
    def idx_hierarchy(self):
        warnings.warn(
            "idx_hierarchy is deprecated, use idx[-1] instead", DeprecationWarning
        )
        return self.idx[-1]

    def show(self, *args, fullscreen=False, **kwargs):
        """Show the mesh (see plot())."""
        import matplotlib.pyplot as plt

        self.plot(*args, **kwargs)
        if fullscreen:
            mng = plt.get_current_fig_manager()
            # mng.frame.Maximize(True)
            mng.window.showMaximized()
        plt.show()
        plt.close()

    def save(self, filename, *args, **kwargs):
        """Save the mesh to a file, either as a PNG/SVG or a mesh file"""
        if pathlib.Path(filename).suffix in [".png", ".svg"]:
            import matplotlib.pyplot as plt

            self.plot(*args, **kwargs)
            plt.savefig(filename, transparent=True, bbox_inches="tight")
            plt.close()
        else:
            self.write(filename)

    def plot(self, *args, **kwargs):
        if self.n == 2:
            self._plot_line(*args, **kwargs)
        else:
            assert self.n == 3
            self._plot_tri(*args, **kwargs)

    def _plot_line(self):
        import matplotlib.pyplot as plt

        if len(self.points.shape) == 1:
            x = self.points
            y = np.zeros(self.points.shape[0])
        else:
            assert len(self.points.shape) == 2 and self.points.shape[1] == 2
            x, y = self.points.T

        plt.plot(x, y, "-o")

    def _plot_tri(
        self,
        show_coedges=True,
        control_volume_centroid_color=None,
        mesh_color="k",
        nondelaunay_edge_color=None,
        boundary_edge_color=None,
        comesh_color=(0.8, 0.8, 0.8),
        show_axes=True,
        cell_quality_coloring=None,
        show_point_numbers=False,
        show_edge_numbers=False,
        show_cell_numbers=False,
        cell_mask=None,
        mark_points=None,
        mark_edges=None,
        mark_cells=None,
    ):
        """Show the mesh using matplotlib."""
        # Importing matplotlib takes a while, so don't do that at the header.
        import matplotlib.pyplot as plt
        from matplotlib.collections import LineCollection, PatchCollection
        from matplotlib.patches import Polygon

        fig = plt.figure()
        ax = fig.gca()
        plt.axis("equal")
        if not show_axes:
            ax.set_axis_off()

        xmin = np.amin(self.points[:, 0])
        xmax = np.amax(self.points[:, 0])
        ymin = np.amin(self.points[:, 1])
        ymax = np.amax(self.points[:, 1])

        width = xmax - xmin
        xmin -= 0.1 * width
        xmax += 0.1 * width

        height = ymax - ymin
        ymin -= 0.1 * height
        ymax += 0.1 * height

        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        # for k, x in enumerate(self.points):
        #     if self.is_boundary_point[k]:
        #         plt.plot(x[0], x[1], "g.")
        #     else:
        #         plt.plot(x[0], x[1], "r.")

        if show_point_numbers:
            for i, x in enumerate(self.points):
                plt.text(
                    x[0],
                    x[1],
                    str(i),
                    bbox={"facecolor": "w", "alpha": 0.7},
                    horizontalalignment="center",
                    verticalalignment="center",
                )

        if show_edge_numbers:
            if self.edges is None:
                self.create_facets()
            for i, point_ids in enumerate(self.edges["points"]):
                midpoint = np.sum(self.points[point_ids], axis=0) / 2
                plt.text(
                    midpoint[0],
                    midpoint[1],
                    str(i),
                    bbox={"facecolor": "b", "alpha": 0.7},
                    color="w",
                    horizontalalignment="center",
                    verticalalignment="center",
                )

        if show_cell_numbers:
            for i, x in enumerate(self.cell_centroids):
                plt.text(
                    x[0],
                    x[1],
                    str(i),
                    bbox={"facecolor": "r", "alpha": 0.5},
                    horizontalalignment="center",
                    verticalalignment="center",
                )

        # coloring
        if cell_quality_coloring:
            cmap, cmin, cmax, show_colorbar = cell_quality_coloring
            plt.tripcolor(
                self.points[:, 0],
                self.points[:, 1],
                self.cells("points"),
                self.q_radius_ratio,
                shading="flat",
                cmap=cmap,
                vmin=cmin,
                vmax=cmax,
            )
            if show_colorbar:
                plt.colorbar()

        if mark_points is not None:
            idx = mark_points
            plt.plot(self.points[idx, 0], self.points[idx, 1], "x", color="r")

        if mark_cells is not None:
            if np.asarray(mark_cells).dtype == bool:
                mark_cells = np.where(mark_cells)[0]

            patches = [
                Polygon(self.points[self.cells("points")[idx]]) for idx in mark_cells
            ]
            p = PatchCollection(patches, facecolor="C1")
            ax.add_collection(p)

        if self.edges is None:
            self.create_facets()

        # Get edges, cut off z-component.
        e = self.points[self.edges["points"]][:, :, :2]

        if nondelaunay_edge_color is None:
            line_segments0 = LineCollection(e, color=mesh_color)
            ax.add_collection(line_segments0)
        else:
            # Plot regular edges, mark those with negative ce-ratio red.
            is_pos = np.zeros(len(self.edges["points"]), dtype=bool)
            is_pos[self.interior_facets[self.signed_circumcenter_distances >= 0]] = True

            # Mark Delaunay-conforming boundary edges
            is_pos_boundary = self.ce_ratios[self.is_boundary_facet_local] >= 0
            is_pos[self.boundary_facets[is_pos_boundary]] = True

            line_segments0 = LineCollection(e[is_pos], color=mesh_color)
            ax.add_collection(line_segments0)
            #
            line_segments1 = LineCollection(e[~is_pos], color=nondelaunay_edge_color)
            ax.add_collection(line_segments1)

        if mark_edges is not None:
            e = self.points[self.edges["points"][mark_edges]][..., :2]
            ax.add_collection(LineCollection(e, color="r"))

        if show_coedges:
            # Connect all cell circumcenters with the edge midpoints
            cc = self.cell_circumcenters

            edge_midpoints = 0.5 * (
                self.points[self.edges["points"][:, 0]]
                + self.points[self.edges["points"][:, 1]]
            )

            # Plot connection of the circumcenter to the midpoint of all three
            # axes.
            a = np.stack(
                [cc[:, :2], edge_midpoints[self.cells("edges")[:, 0], :2]], axis=1
            )
            b = np.stack(
                [cc[:, :2], edge_midpoints[self.cells("edges")[:, 1], :2]], axis=1
            )
            c = np.stack(
                [cc[:, :2], edge_midpoints[self.cells("edges")[:, 2], :2]], axis=1
            )

            line_segments = LineCollection(
                np.concatenate([a, b, c]), color=comesh_color
            )
            ax.add_collection(line_segments)

        if boundary_edge_color:
            e = self.points[self.edges["points"][self.is_boundary_facet]][:, :, :2]
            line_segments1 = LineCollection(e, color=boundary_edge_color)
            ax.add_collection(line_segments1)

        if control_volume_centroid_color is not None:
            centroids = self.get_control_volume_centroids(cell_mask=cell_mask)
            ax.plot(
                centroids[:, 0],
                centroids[:, 1],
                linestyle="",
                marker=".",
                color=control_volume_centroid_color,
            )
            for k, centroid in enumerate(centroids):
                plt.text(
                    centroid[0],
                    centroid[1],
                    str(k),
                    bbox=dict(facecolor=control_volume_centroid_color, alpha=0.7),
                    horizontalalignment="center",
                    verticalalignment="center",
                )

        return fig



import pathlib

import meshio
import numpy as np
import pytest

import meshplex

this_dir = pathlib.Path(__file__).resolve().parent


@pytest.mark.parametrize(
    "points,cells,ref",
    [
        # line
        ([[0.0], [0.35]], [[0, 1]], [0.35]),
        ([[0.0], [0.35]], [[1, 0]], [-0.35]),
        # triangle
        ([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]], [[0, 1, 2]], [0.5]),
        ([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0]], [[0, 1, 2]], [-0.5]),
        (
            [[0.0, 0.0], [1.0, 0.0], [1.1, 1.0], [0.0, 1.0]],
            [[0, 1, 2], [0, 3, 2]],
            [0.5, -0.55],
        ),
        # tetra
        (
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
            [[0, 1, 2, 3]],
            [1 / 6],
        ),
        (
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
            [[0, 1, 3, 2]],
            [-1 / 6],
        ),
    ],
)
def test_signed_area(points, cells, ref):
    mesh = meshplex.Mesh(points, cells)
    ref = np.array(ref)
    assert mesh.signed_cell_volumes.shape == ref.shape
    assert np.all(
        np.abs(ref - mesh.signed_cell_volumes) < np.abs(ref) * 1.0e-13 + 1.0e-13
    )


def test_signed_area_pacman():
    mesh = meshio.read(this_dir / "meshes" / "pacman.vtu")
    assert np.all(np.abs(mesh.points[:, 2]) < 1.0e-15)
    X = mesh.points[:, :2]

    mesh = meshplex.Mesh(X, mesh.get_cells_type("triangle"))

    vols = mesh.signed_cell_volumes
    # all cells are positively oriented in this mesh
    assert np.all(mesh.signed_cell_volumes > 0.0)
    assert np.all(abs(abs(vols) - mesh.cell_volumes) < 1.0e-12 * mesh.cell_volumes)



import pathlib

import numpy as np
import pytest

import meshplex

from .mesh_tri.helpers import assert_mesh_consistency, compute_all_entities


def get_mesh0():
    # _____________
    # |   _/ \_   |
    # | _/     \_ |
    # |/_________\|
    # |\_       _/|
    # |  \_   _/  |
    # |___ \_/____|
    #
    points = [
        [0.0, 0.0],
        [1.0, 0.0],
        [1.0, 1.0],
        [0.0, 1.0],
        [0.0, 0.5],
        [0.5, 0.0],
        [1.0, 0.5],
        [0.5, 1.0],
    ]
    cells = [
        [0, 5, 4],
        [5, 1, 6],
        [6, 2, 7],
        [7, 3, 4],
        [5, 6, 4],
        [6, 7, 4],
    ]
    mesh = meshplex.Mesh(points, cells)
    mesh.create_facets()
    return mesh


def get_mesh1():
    # _____________
    # |\_       _/|
    # |  \_   _/  |
    # |    \_/    |
    # |   _/ \_   |
    # | _/     \_ |
    # |/_________\|
    #
    points = [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0], [0.5, 0.5]]
    cells = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [0, 4, 3]]
    return meshplex.Mesh(points, cells)


def get_mesh2():
    this_dir = pathlib.Path(__file__).resolve().parent
    mesh0 = meshplex.read(this_dir / "meshes" / "pacman.vtu")
    return meshplex.Mesh(mesh0.points[:, :2], mesh0.cells("points"))


@pytest.mark.parametrize(
    "remove_idx,expected_num_cells,expected_num_edges",
    [
        # remove corner cell
        [[0], 5, 11],
        # remove corner cell
        [[0, 4], 4, 10],
        # remove interior cells
        [[4, 5], 4, 12],
        # remove no cells at all
        [[], 6, 13],
    ],
)
def test_remove_cells(remove_idx, expected_num_cells, expected_num_edges):
    mesh = get_mesh0()
    assert len(mesh.cells("points")) == 6
    assert len(mesh.edges["points"]) == 13
    # remove a corner cell
    mesh.remove_cells(remove_idx)
    assert len(mesh.cells("points")) == expected_num_cells
    assert len(mesh.edges["points"]) == expected_num_edges
    assert_mesh_consistency(mesh)


def test_remove_cells_boundary():
    mesh = get_mesh1()

    assert np.all(mesh.is_boundary_point == [True, True, True, True, False])
    assert np.all(mesh.is_boundary_facet_local[0] == [False, False, False, False])
    assert np.all(mesh.is_boundary_facet_local[1] == [False, False, False, True])
    assert np.all(mesh.is_boundary_facet_local[2] == [True, True, True, False])
    assert np.all(
        mesh.is_boundary_facet == [True, True, False, True, False, True, False, False]
    )
    assert np.all(mesh.is_boundary_cell)
    assert np.all(mesh.facets_cells_idx == [0, 1, 0, 2, 1, 3, 2, 3])
    # cell id:
    assert np.all(mesh.facets_cells["boundary"][1] == [0, 3, 1, 2])
    # local edge:
    assert np.all(mesh.facets_cells["boundary"][2] == [2, 1, 2, 2])
    # cell id:
    assert np.all(
        mesh.facets_cells["interior"][1:3].T == [[0, 3], [0, 1], [1, 2], [2, 3]]
    )
    # local edge:
    assert np.all(
        mesh.facets_cells["interior"][3:5].T == [[1, 2], [0, 1], [0, 1], [0, 0]]
    )

    # now lets remove some cells
    mesh.remove_cells([0])

    assert_mesh_consistency(mesh)

    assert np.all(mesh.is_boundary_point)
    assert np.all(mesh.is_boundary_facet_local[0] == [False, False, False])
    assert np.all(mesh.is_boundary_facet_local[1] == [True, False, True])
    assert np.all(mesh.is_boundary_facet_local[2] == [True, True, True])
    assert np.all(
        mesh.is_boundary_facet == [True, True, True, True, True, False, False]
    )
    assert np.all(mesh.is_boundary_cell)


def test_remove_boundary_cell():
    mesh = get_mesh0()
    mesh.remove_boundary_cells(
        lambda ibc: np.all(mesh.cell_centroids[ibc] > 0.5, axis=1)
    )
    assert mesh.cells("points").shape[0] == 5


def test_remove_all():
    points = [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
    cells = [[0, 1, 2]]
    mesh = meshplex.Mesh(points, cells)
    assert np.all(mesh.is_point_used)

    mesh.remove_cells([0])
    assert_mesh_consistency(mesh)
    assert not np.any(mesh.is_point_used)


@pytest.mark.parametrize(
    "mesh0, remove_cells",
    [
        (get_mesh0(), [0]),
        (get_mesh1(), [0, 1]),
        (get_mesh2(), [0, 3, 57, 59, 60, 61, 100]),
    ],
)
def test_reference(mesh0, remove_cells):
    # some dummy calls to make sure the respective value are computed before the cell
    # removal and then updated
    compute_all_entities(mesh0)

    # now remove some cells
    mesh0.remove_cells(remove_cells)

    assert_mesh_consistency(mesh0)


def test_remove_duplicate():
    # lines
    points = [0.0, 0.1, 0.7, 1.0]
    cells = [[0, 1], [1, 2], [2, 1], [0, 1], [3, 2]]
    mesh = meshplex.Mesh(points, cells)
    n = mesh.remove_duplicate_cells()
    assert n == 2
    assert np.all(mesh.cells("points") == [[0, 1], [1, 2], [3, 2]])

    # triangle
    points = [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0]]
    cells = [[0, 2, 3], [0, 1, 2], [0, 2, 1]]
    mesh = meshplex.Mesh(points, cells)
    n = mesh.remove_duplicate_cells()
    assert n == 1
    assert np.all(mesh.cells("points") == [[0, 2, 3], [0, 1, 2]])

    # tetrahedra
    points = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    cells = [[0, 1, 2, 3], [3, 1, 2, 0]]
    mesh = meshplex.Mesh(points, cells)
    n = mesh.remove_duplicate_cells()
    assert n == 1
    assert np.all(mesh.cells("points") == [[0, 1, 2, 3]])


if __name__ == "__main__":
    test_remove_cells_boundary()



import numpy as np

import meshplex

from .helpers import assert_mesh_consistency, compute_all_entities


def test():
    points = [
        [-2.1, -3.1],
        [0.0, 0.0],
        [1.0, 0.0],
        [-2.1, -3.1],
        [1.0, 1.0],
        [0.0, 1.0],
    ]
    cells = [[1, 2, 4], [1, 4, 5]]
    mesh = meshplex.MeshTri(points, cells)
    compute_all_entities(mesh)

    mesh.remove_dangling_points()

    assert len(mesh.points) == 4
    assert np.all(mesh.cells("points") == np.array([[0, 1, 2], [0, 2, 3]]))
    assert np.all(
        mesh.edges["points"] == np.array([[0, 1], [0, 2], [0, 3], [1, 2], [2, 3]])
    )
    assert_mesh_consistency(mesh)



import numpy as np

from ._mesh import Mesh

__all__ = ["MeshTetra"]


class MeshTetra(Mesh):
    """Class for handling tetrahedral meshes."""

    def __init__(self, points, cells, sort_cells=False):
        super().__init__(points, cells, sort_cells=sort_cells)
        assert self.n == 4
        self.faces = None

    def _create_face_edge_relationships(self):
        a = np.vstack(
            [
                self.faces["points"][:, [1, 2]],
                self.faces["points"][:, [2, 0]],
                self.faces["points"][:, [0, 1]],
            ]
        )

        # Find the unique edges
        b = np.ascontiguousarray(a).view(
            np.dtype((np.void, a.dtype.itemsize * a.shape[1]))
        )
        _, idx, inv = np.unique(b, return_index=True, return_inverse=True)
        edge_points = a[idx]

        self.edges = {"points": edge_points}

        # face->edge relationship
        num_faces = len(self.faces["points"])
        face_edges = inv.reshape([3, num_faces]).T
        self.faces["edges"] = face_edges

    @property
    def q_min_sin_dihedral_angles(self):
        """Get the smallest of the sines of the 6 angles between the faces of each
        tetrahedron, times a scaling factor that makes sure the value is 1 for the
        equilateral tetrahedron.
        """
        # https://math.stackexchange.com/a/49340/36678
        fa = self.facet_areas

        el2 = self.ei_dot_ei
        a = el2[0][0]
        b = el2[1][0]
        c = el2[2][0]
        d = el2[0][2]
        e = el2[1][1]
        f = el2[0][1]

        cos_alpha = []

        H2 = (4 * a * d - ((b + e) - (c + f)) ** 2) / 16
        J2 = (4 * b * e - ((a + d) - (c + f)) ** 2) / 16
        K2 = (4 * c * f - ((a + d) - (b + e)) ** 2) / 16

        # Angle between face 0 and face 1.
        # The faces share (face 0, edge 0), (face 1, edge 2).
        cos_alpha += [(fa[0] ** 2 + fa[1] ** 2 - H2) / (2 * fa[0] * fa[1])]
        # Angle between face 0 and face 2.
        # The faces share (face 0, edge 1), (face 2, edge 1).
        cos_alpha += [(fa[0] ** 2 + fa[2] ** 2 - J2) / (2 * fa[0] * fa[2])]
        # Angle between face 0 and face 3.
        # The faces share (face 0, edge 2), (face 3, edge 0).
        cos_alpha += [(fa[0] ** 2 + fa[3] ** 2 - K2) / (2 * fa[0] * fa[3])]
        # Angle between face 1 and face 2.
        # The faces share (face 1, edge 0), (face 2, edge 2).
        cos_alpha += [(fa[1] ** 2 + fa[2] ** 2 - K2) / (2 * fa[1] * fa[2])]
        # Angle between face 1 and face 3.
        # The faces share (face 1, edge 1), (face 3, edge 1).
        cos_alpha += [(fa[1] ** 2 + fa[3] ** 2 - J2) / (2 * fa[1] * fa[3])]
        # Angle between face 2 and face 3.
        # The faces share (face 2, edge 0), (face 3, edge 2).
        cos_alpha += [(fa[2] ** 2 + fa[3] ** 2 - H2) / (2 * fa[2] * fa[3])]

        cos_alpha = np.array(cos_alpha).T
        sin_alpha = np.sqrt(1 - cos_alpha**2)

        m = np.min(sin_alpha, axis=1) / (np.sqrt(2) * 2 / 3)
        return m

    @property
    def q_vol_rms_edgelength3(self):
        """For each cell, return the ratio of the volume and the cube of the
        root-mean-square edge length. (This is cell quality measure used by Stellar
        <https://people.eecs.berkeley.edu/~jrs/stellar/>.)
        """
        el2 = self.ei_dot_ei
        rms = np.sqrt(
            (el2[0][0] + el2[1][0] + el2[2][0] + el2[0][2] + el2[1][1] + el2[0][1]) / 6
        )
        alpha = np.sqrt(2) / 12  # normalization factor
        return self.cell_volumes / rms**3 / alpha

    def show(self):
        from matplotlib import pyplot as plt

        self.plot()
        plt.show()
        plt.close()

    def plot(self):
        from matplotlib import pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        ax = plt.axes(projection=Axes3D.name)
        # "It is not currently possible to manually set the aspect on 3D axes"
        # plt.axis("equal")

        for cell_id in range(len(self.cells("points"))):
            cc = self.cell_circumcenters[cell_id]
            #
            face_ccs = self._circumcenters[-2]

            # draw the face circumcenters
            ax.plot(
                face_ccs[..., 0].flatten(),
                face_ccs[..., 1].flatten(),
                face_ccs[..., 2].flatten(),
                "go",
            )
            # draw the connections
            #   tet circumcenter---face circumcenter
            for face_cc in face_ccs:
                ax.plot(
                    [cc[0], face_cc[cell_id, 0]],
                    [cc[1], face_cc[cell_id, 1]],
                    [cc[2], face_cc[cell_id, 2]],
                    "b-",
                )

    def show_edge(self, edge_id):
        from matplotlib import pyplot as plt

        self.plot_edge(edge_id)
        plt.show()
        plt.close()

    def plot_edge(self, edge_id):
        """Displays edge with ce_ratio.

        :param edge_id: Edge ID for which to show the ce_ratio.
        :type edge_id: int
        """
        # pylint: disable=unused-variable,relative-import
        from matplotlib import pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        if self._cells_facets is None:
            self.create_facets()
        if "edges" not in self.faces:
            self._create_face_edge_relationships()

        ax = plt.axes(projection=Axes3D.name)
        # "It is not currently possible to manually set the aspect on 3D axes"
        # plt.axis("equal")

        # find all faces with this edge
        adj_face_ids = np.where((self.faces["edges"] == edge_id).any(axis=1))[0]
        # find all cells with the faces
        # https://stackoverflow.com/a/38481969/353337
        adj_cell_ids = np.where(
            np.in1d(self.cells("facets"), adj_face_ids)
            .reshape(self.cells("facets").shape)
            .any(axis=1)
        )[0]

        # plot all those adjacent cells; first collect all edges
        adj_edge_ids = np.unique(
            [
                adj_edge_id
                for adj_cell_id in adj_cell_ids
                for face_id in self.cells("facets")[adj_cell_id]
                for adj_edge_id in self.faces["edges"][face_id]
            ]
        )
        col = "k"
        for adj_edge_id in adj_edge_ids:
            x = self.points[self.edges["points"][adj_edge_id]]
            ax.plot(x[:, 0], x[:, 1], x[:, 2], col)

        # make clear which is edge_id
        x = self.points[self.edges["points"][edge_id]]
        ax.plot(x[:, 0], x[:, 1], x[:, 2], color=col, linewidth=3.0)

        # connect the face circumcenters with the corresponding cell
        # circumcenters
        for cell_id in adj_cell_ids:
            cc = self.cell_circumcenters[cell_id]
            face_ccs = self._circumcenters[-2]
            # draw the face circumcenters
            ax.plot(
                face_ccs[..., 0].flatten(),
                face_ccs[..., 1].flatten(),
                face_ccs[..., 2].flatten(),
                "go",
            )
            # draw the connections
            #   tet circumcenter---face circumcenter
            for face_cc in face_ccs:
                ax.plot(
                    [cc[0], face_cc[cell_id, 0]],
                    [cc[1], face_cc[cell_id, 1]],
                    [cc[2], face_cc[cell_id, 2]],
                    "b-",
                )

        # draw the cell circumcenters
        cc = self.cell_circumcenters[adj_cell_ids]
        ax.plot(cc[:, 0], cc[:, 1], cc[:, 2], "ro")

    def show_cell(
        self,
        cell_id,
        control_volume_boundaries_rgba=None,
        barycenter_rgba=None,
        circumcenter_rgba=None,
        incenter_rgba=None,
        face_circumcenter_rgba=None,
        insphere_rgba=None,
        circumsphere_rgba=None,
        line_width=1.0,
        close=False,
        render=True,
    ):
        import vtk

        def get_line_actor(x0, x1, line_width=1.0):
            source = vtk.vtkLineSource()
            source.SetPoint1(x0)
            source.SetPoint2(x1)
            # mapper
            mapper = vtk.vtkPolyDataMapper()
            mapper.SetInputConnection(source.GetOutputPort())
            # actor
            actor = vtk.vtkActor()
            actor.SetMapper(mapper)
            # color actor
            actor.GetProperty().SetColor(0, 0, 0)
            actor.GetProperty().SetLineWidth(line_width)
            return actor

        def get_sphere_actor(x0, r, rgba):
            # Generate polygon data for a sphere
            sphere = vtk.vtkSphereSource()

            sphere.SetCenter(x0)
            sphere.SetRadius(r)

            sphere.SetPhiResolution(100)
            sphere.SetThetaResolution(100)

            # Create a mapper for the sphere data
            sphere_mapper = vtk.vtkPolyDataMapper()
            # sphere_mapper.SetInput(sphere.GetOutput())
            sphere_mapper.SetInputConnection(sphere.GetOutputPort())

            # Connect the mapper to an actor
            sphere_actor = vtk.vtkActor()
            sphere_actor.SetMapper(sphere_mapper)
            sphere_actor.GetProperty().SetColor(rgba[:3])
            sphere_actor.GetProperty().SetOpacity(rgba[3])
            return sphere_actor

        # Visualize
        renderer = vtk.vtkRenderer()
        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)

        for ij in [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]:
            x0, x1 = self.points[self.cells("points")[cell_id][ij]]
            renderer.AddActor(get_line_actor(x0, x1, line_width))
        renderer.SetBackground(1.0, 1.0, 1.0)

        r = 0.02

        if circumcenter_rgba is not None:
            renderer.AddActor(
                get_sphere_actor(self.cell_circumcenters[cell_id], r, circumcenter_rgba)
            )

        if circumsphere_rgba is not None:
            renderer.AddActor(
                get_sphere_actor(
                    self.cell_circumcenters[cell_id],
                    self.cell_circumradius[cell_id],
                    circumsphere_rgba,
                )
            )

        if incenter_rgba is not None:
            renderer.AddActor(
                get_sphere_actor(self.cell_incenters[cell_id], r, incenter_rgba)
            )

        if insphere_rgba is not None:
            renderer.AddActor(
                get_sphere_actor(
                    self.cell_incenters[cell_id],
                    self.cell_inradius[cell_id],
                    insphere_rgba,
                )
            )

        if barycenter_rgba is not None:
            renderer.AddActor(
                get_sphere_actor(self.cell_barycenters[cell_id], r, barycenter_rgba)
            )

        if face_circumcenter_rgba is not None:
            face_ccs = self._circumcenters[-2][:, 0]
            for f in face_ccs:
                renderer.AddActor(get_sphere_actor(f, r, face_circumcenter_rgba))

        if control_volume_boundaries_rgba:
            cell_cc = self.cell_circumcenters[cell_id]
            face_ccs = self._circumcenters[-2][:, 0]
            for face, face_cc in zip(range(4), face_ccs):
                for edge in range(3):
                    k0, k1 = self.idx[-1][:, edge, face, cell_id]
                    edge_midpoint = 0.5 * (self.points[k0] + self.points[k1])

                    points = vtk.vtkPoints()
                    points.InsertNextPoint(*edge_midpoint)
                    points.InsertNextPoint(*cell_cc)
                    points.InsertNextPoint(*face_cc)

                    triangle = vtk.vtkTriangle()
                    triangle.GetPointIds().SetId(0, 0)
                    triangle.GetPointIds().SetId(1, 1)
                    triangle.GetPointIds().SetId(2, 2)

                    triangles = vtk.vtkCellArray()
                    triangles.InsertNextCell(triangle)

                    trianglePolyData = vtk.vtkPolyData()
                    trianglePolyData.SetPoints(points)
                    trianglePolyData.SetPolys(triangles)

                    # mapper
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(trianglePolyData)

                    # actor
                    actor = vtk.vtkActor()
                    actor.SetMapper(mapper)

                    actor.GetProperty().SetColor(*control_volume_boundaries_rgba[:3])
                    actor.GetProperty().SetOpacity(control_volume_boundaries_rgba[3])
                    renderer.AddActor(actor)

        if render:
            render_window.Render()

        if close:
            render_window.Finalize()
            del render_window, render_window_interactor
        else:
            render_window_interactor.Start()



import meshio
import numpy as np

from ._mesh_tetra import MeshTetra
from ._mesh_tri import MeshTri

__all__ = ["read"]


def _sanitize(points, cells):
    uvertices, uidx = np.unique(cells, return_inverse=True)
    cells = uidx.reshape(cells.shape)
    points = points[uvertices]
    return points, cells


def from_meshio(mesh):
    """Transform from meshio to meshplex format.

    :param mesh: The meshio mesh object.
    :type mesh: meshio.Mesh
    :returns mesh{2,3}d: The mesh data.
    """
    # make sure to include the used nodes only
    tetra = mesh.get_cells_type("tetra")
    if len(tetra) > 0:
        points, cells = _sanitize(mesh.points, tetra)
        return MeshTetra(points, cells)

    tri = mesh.get_cells_type("triangle")
    assert len(tri) > 0
    points, cells = _sanitize(mesh.points, tri)
    return MeshTri(points, cells)


def read(filename):
    """Reads an unstructured mesh into meshplex format.

    :param filenames: The files to read from.
    :type filenames: str
    :returns mesh{2,3}d: The mesh data.
    """
    return from_meshio(meshio.read(filename))



import numpy as np

import meshplex


def test_degenerate_cell(tol=1.0e-14):
    points = [[0.0, 0.0], [0.5, 0.0], [1.0, 0.0]]
    cells = [[0, 1, 2]]
    mesh = meshplex.Mesh(points, cells)

    ref = np.array([0.0])
    assert np.all(np.abs(mesh.cell_volumes - ref) < tol * (1.0 + ref))

    ref = np.array([[0.0], [0.0], [0.0]])
    assert np.all(np.abs(mesh.cell_heights - ref) < tol * (1.0 + ref))

    # inf, not nan
    assert np.all(mesh.cell_circumradius == [np.inf])
    assert np.all(mesh.circumcenter_facet_distances == [[np.inf], [-np.inf], [np.inf]])
    assert np.all(
        mesh.cell_partitions
        == [[[np.inf], [-np.inf], [np.inf]], [[np.inf], [-np.inf], [np.inf]]]
    )

    # those are nan
    assert np.all(np.isnan(mesh.cell_circumcenters))


def test_repeated_point(tol=1.0e-14):
    """Special case of the degeneracy: A point is doubled up."""
    points = [
        [0.0, 0.0],
        [1.0, 0.0],
    ]
    cells = [[0, 1, 1]]
    mesh = meshplex.Mesh(points, cells)

    ref = np.array([0.0])
    assert np.all(np.abs(mesh.cell_volumes - ref) < tol * (1.0 + ref))

    ref = np.array([[1.0], [0.0], [0.0]])
    assert np.all(np.abs(mesh.cell_heights - ref) < tol * (1.0 + ref))

    ref = np.array([0.5])
    assert np.all(np.abs(mesh.cell_circumradius - ref) < tol * (1.0 + ref))

    ref = np.array([[0.5, 0.0]])
    assert np.all(np.abs(mesh.cell_circumcenters - ref) < tol * (1.0 + ref))

    # TODO
    # print(f"{mesh.circumcenter_facet_distances = }")
    # assert np.all(mesh.circumcenter_facet_distances == [[0.5], [0.0], [0.0]])
    # print(mesh.cell_partitions)
    # assert np.all(
    #     mesh.cell_partitions
    #     == [[[np.inf], [-np.inf], [np.inf]], [[np.inf], [-np.inf], [np.inf]]]
    # )


def test_all_same_point(tol=1.0e-14):
    """Even more extreme: A simplex made of one point."""
    points = [[0.0, 0.0]]
    cells = [[0, 0, 0]]
    mesh = meshplex.Mesh(points, cells)

    ref = np.array([0.0])
    assert np.all(np.abs(mesh.cell_volumes - ref) < tol * (1.0 + ref))

    ref = np.array([[0.0], [0.0], [0.0]])
    assert np.all(np.abs(mesh.cell_volumes - ref) < tol * (1.0 + ref))

    # TODO for some reason, this test is unstable on gh-actions
    # ref = np.array([[0.0], [0.0], [0.0]])
    # assert np.all(np.abs(mesh.cell_heights - ref) < tol * (1.0 + ref))

    # TODO
    # assert np.all(mesh.cell_circumradius == [0.0])
    # assert np.all(mesh.circumcenter_facet_distances == [[0.0], [0.0], [0.0]])
    # assert np.all(
    #     mesh.cell_partitions
    #     == [[[np.inf], [-np.inf], [np.inf]], [[np.inf], [-np.inf], [np.inf]]]
    # )
    # ref = np.array([[0.0, 0.0]])
    # assert np.all(np.abs(mesh.cell_circumcenters - ref) < tol * (1.0 + ref))


def test_degenerate_flip():
    # almost degenerate
    points = [
        [0.0, 0.0],
        [0.5, -1.0e-5],
        [1.0, 0.0],
        [0.5, 0.5],
    ]
    cells = [[0, 2, 1], [0, 2, 3]]
    mesh = meshplex.MeshTri(points, cells)
    num_flips = mesh.flip_until_delaunay()
    assert num_flips == 1
    ref = np.array([[1, 0, 3], [1, 3, 2]])
    assert np.all(mesh.cells("points") == ref)

    # make sure the same thing happens if the cell is exactly degenerate
    points = [
        [0.0, 0.0],
        [0.5, 0.0],
        [1.0, 0.0],
        [0.5, 0.5],
    ]
    cells = [[0, 2, 1], [0, 2, 3]]
    mesh = meshplex.MeshTri(points, cells)
    num_flips = mesh.flip_until_delaunay()
    assert num_flips == 1
    ref = np.array([[1, 0, 3], [1, 3, 2]])
    assert np.all(mesh.cells("points") == ref)
